# -*- coding: utf-8 -*-

# This file is part of pyChemEngg python package.
 
# PyChemEngg: A python-based framework to promote problem solving and critical
# thinking in chemical engineering.

# Copyright (c) 2021 Harvinder Singh Gill <profhsgill@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Contains correlations to find Nusselt number
for INTERNAL flow

"""
import numpy as np
from scipy import interpolate
from scipy.optimize import fsolve
import warnings
warnings.simplefilter('always')

__all__ = ["CircularTube"]

class CircularTube:
    r""" For heat transfer from fluid flow through a circular tube.


    Parameters
    ----------
    `None_required` : 'None'
        This class takes no parameters for instance creation.
 
    
    Attributes
    ----------
    `None_required` : 'None'
        This class does not expose any instance attributes.
    

    Examples
    --------
    First import the module **internalflow**.
       
    >>> from pychemengg.heattransfer import internalflow as intflow 
    >>> tube = intflow.CircularTube
    # This will assign the class 'CircularTube' to the variable 'tube'.
    # Methods of the class 'CircularTube' can then be called like so :-
    # tube.method(kwarg1=x, ...)

    """
    
    def Nu_uniformheatflux_laminar(self):
        r""" Nusselt number for laminar flow in tube with constant heat flux.
        
        
        Parameters
        ----------
        `None_required` : 'None'
        

        Returns
        -------
        Nu : `float`
            Nusselt number = 4.36
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 4.36
    

        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_uniformheatflux_laminar()
        4.36
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        """
        return 4.36
    
    def Nu_constantsurfacetemp_laminar(self):
        r""" Nusselt number for laminar flow in tube (constant tube surface
        temperature).
        
        
        Parameters
        ----------
        `None_required` : 'None'
        

        Returns
        -------
        Nu : `float`
            Nusselt number = 3.36
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 3.36
        

        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_constantsurfacetemp_laminar()
        3.36
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        return 3.66
    
    def Nu_thermallydeveloping_laminar_edwards(self, Re=None, Pr=None,
                                               length=None, diameter=None):
        r"""Nusselt number for thermally developing laminar flow and
        constant tube surface temperature.
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number for tube/pipe
        Pr : `int or float`
            Prandtl number for the fluid.
        length : `int or float`
            Length of tube
        diameter : `int or float`
            Diameter of tube
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 3.66 + \cfrac {0.065 (D/L) Re Pr}
                               {1 + 0.04[(D/L) Re Pr]^{2/3}}
            
               
        *where:*
        
            Re = Reynolds number
            
            Pr = Prandtl number
            
            D = Pipe inner diameter
            
            L = Pipe length
            
            :math:`Re < 2300` (for laminar flow)
            
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                
                except :math:`\mu_{surface}`
                

        
        Warnings
        --------
        A Nusselt number is returned based on the equation
        even if parameters (such as *Re*, *Pr*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
                

        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_thermallydeveloping_laminar_edwards(Re=77.16, Pr=28750,
                                              length=1500, diameter=0.4)
        13.729061335098443

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        """
        term1 = diameter/length*Re*Pr
        Nu = 3.66 + (0.065*term1)/(1+0.04*np.power(term1, 2/3))
        if not (Re < 2300): warnings.warn("Flow is not laminar. Re is not within range. ")
        return Nu
    
    def Nu_thermallyandhydrodynamicallydeveloping_laminar_siedertate(self, Re=None, Pr=None, length=None,
                                        diameter=None, viscosity_bulk=None, viscosity_surface=None):
        r"""Nusselt number for thermally developing and hydrodynamically 
        developing laminar flow and constant tube surface temperature.
        
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number for tube/pipe
        Pr : `int or float`
            Prandtl number for the fluid.
        length : `int or float`
            Length of tube
        diameter : `int or float`
            Diameter of tube
        viscosity_bulk : `int or float`
            fluid viscosity at bulk temperature
        viscosity_surface : `int or float`
            fluid viscosity at surface temperature

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 1.86 \left(\frac{Re Pr D} {L}\right)^{1/3} 
                      \left(\frac{\mu_{bulk}} {\mu_{surface}}\right)^{0.14}
            
               
        *where:*
            Re = Reynolds number
            
            Pr = Prandtl number
            
            D = Pipe inner diameter
            
            L = Pipe length
            
            :math:`\mu_{bulk}` = viscosity at bulk temperature
            
            :math:`\mu_{surface}` = viscosity at surface temperature
            
            :math:`Re < 2300` (for laminar flow)
            
            :math:`0.6 \eqslantless Pr \eqslantless 5`
            
            :math:`0.0044 \eqslantless \left(\frac{\mu_{bulk}} {\mu_{surface}}\right) \eqslantless 9.75`
            
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                
                except :math:`\mu_{surface}` is at surface temperature
                

        
        Warnings
        --------
        A Nusselt number is returned based on the equation
        even if parameters (such as *Re*, *Pr*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_thermallyandhydrodynamicallydeveloping_laminar_siedertate(Re=1390, Pr=0.7228,
                                        length=0.1, diameter=0.005, viscosity_bulk=1.963,
                                        viscosity_surface=2.420)
        6.664822962495227

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.        
        """
        Nu = 1.86 * np.power(Re*Pr*diameter/length, 1/3) * np.power(viscosity_bulk/viscosity_surface, 0.14)
        if not (Re < 2300): warnings.warn("Flow is not laminar. Re is not within range. ")
        if not (0.6<=Pr<=5): warnings.warn("Pr is not within range.")
        if not (0.0044<=(viscosity_bulk/viscosity_surface)<=9.75):
                warnings.warn("Ratio of viscosity_bulk to viscosity_surface"
                              " is not within range.")
        return Nu
    
    def Nu_dittusboelter(self, Re=None, Pr=None, thermal_ID="heated"):
        r"""Nusselt number for turbulent flow constant tube surface temperature.
        
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number for tube/pipe
        Pr : `int or float`
            Prandtl number for the fluid.
        thermal_ID : `str`
            Identifier for whether fluid is being "heated" or "cooled".
            Default = "heated"
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 0.023 Re^{0.8} Pr^{n}
        
        *where:*
        
            n = 0.3 when fluid is "cooled"
            
              = 0.4 when fluid is "heated"
            
            Re = Reynolds number
            
            Pr = Prandtl number
            
            :math:`Re > 10000` (for turbulent flow)
            
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                

        
        Warnings
        --------
        A Nusselt number is returned based on the equation
        even if parameters (such as *Re*, *Pr*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_dittusboelter(Re=10760, Pr=4.32, thermal_ID="heated")
        69.40186669836865

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.        
        """
        if thermal_ID == "heated":
            Nu = 0.023 * np.power(Re, 0.8) * np.power(Pr, 0.4)
        if thermal_ID == "cooled":
            Nu = 0.023 * np.power(Re, 0.8) * np.power(Pr, 0.3)
        if not (Re>=10000): warnings.warn("Flow is not turbulent. Re is not within range.")
        return Nu
        
    def Nu_siedertate(self, Re=None, Pr=None, viscosity_bulk=None, viscosity_surfaceace=None):
        r"""Nusselt number for turbulent flow constant tube surface temperature.
        Use when temperature differences are large.
        
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number for tube/pipe
        Pr : `int or float`
            Prandtl number for the fluid.
        viscosity_bulk : `int or float`
            fluid viscosity at bulk temperature
        viscosity_surface : `int or float`
            fluid viscosity at surface temperature
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math:: 
            Nu = 0.027 Re^{0.8} Pr^{1/3} \left( \frac {\mu_{bulk}} {\mu_{surface}} \right)^{0.14}
        
        *where:*
            Re = Reynolds number
            
            Pr = Prandtl number
            
            :math:`\mu_{bulk}` = viscosity at bulk temperature
            
            :math:`\mu_{surface}` = viscosity at surface temperature
            
            :math:`Re > 10000` (for turbulent flow)
            
            :math:`0.7 \eqslantless Pr \eqslantless 16700`
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                
                except :math:`\mu_{surface}` is at surface temperature
                
        
        Warnings
        --------
        A Nusselt number is returned based on the equation
        even if parameters (such as *Re*, *Pr*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_siedertate(self, Re=2.04e5, Pr=3.02,
                               viscosity_bulk=4.71,
                               viscosity_surfaceace=2.82)
        741.7512113817678

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        [2] M. N. Ozisik, "Heat Transfer A Basic Approach", McGraw Hill, 1985.
        """
        Nu = 0.027 * np.power(Re, 0.8) * np.power(Pr, 1/3) * np.power(viscosity_bulk/viscosity_surfaceace, 0.14)
        if not (Re>=10000): warnings.warn("Flow is not turbulent. Re is not within range.")
        if not (0.7<=Pr<=16700): warnings.warn("Pr is not within range.")
        return Nu
     
    def Nu_gnielinski(self, Re=None, Pr=None):
        r"""Nusselt number for turbulent flow constant tube surface temperature.
        A more accurate correlation.
        
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number for tube/pipe
        Pr : `int or float`
            Prandtl number for the fluid.
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = \cfrac{(f/8) (Re-1000) Pr}{ 1+12.7(f/8)^{0.5} (Pr^{2/3}-1)}
            
        
        *where:*
            Re = Reynolds number
            
            Pr = Prandtl number
            
            f = friction factor (can be determined from different correlations.
                Here following 'Petukhov' correlation is used.)
            
            :math:`f=(0.790 \ln Re - 1.64)^{-2}` (Petukhov correlation)
            
            :math:`3*10^3\eqslantless Re \eqslantless 5*10^6` (for turbulent flow)
            
            :math:`0.5 \eqslantless Pr \eqslantless 2000`
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                
        
        Warnings
        --------
        A Nusselt number is returned based on the equation
        even if parameters (such as *Re*, *Pr*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_siedertate(Re=2.04e5, Pr=3.02,
                               viscosity_bulk=4.71,
                               viscosity_surfaceace=2.82)
        741.7512113817678

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        [2] M. N. Ozisik, "Heat Transfer A Basic Approach", McGraw Hill, 1985.
        """
        f = np.power(0.79*np.log(Re) - 1.64, -2)
        num = (f/8)*(Re-1000)*Pr
        denom = 1 + 12.7*np.power(f/8, 0.5)*(np.power(Pr,2/3)-1)
        Nu = num/denom
        if not (3e3<=Re<=5e6): warnings.warn("Re is not within range.")
        if not (0.5<=Pr<=2000): warnings.warn("Pr is not within range.")
        return Nu

    def Nu_annulus_laminar(self, outerannulus_diameter=None,
                           innerannulus_diameter=None,
                           findNUat_ID="innerannulus_diameter"):
        
        r"""Nusselt number for laminar flow in annulus (one surface is
        isothermal and other adiabatic)
        
        
        Parameters
        ----------
        outerannulus_diameter : `int or float`
            Outer annulus diameter for fluid flow.
        innerannulus_diameter : `int or float`
            Inner annulus diameter for fluid flow.
        findNUat_ID : `str`
            Annulus diameter where  Nu is to be found. 
            Default = "innerannulus_diameter", 
            second option = "outerannulus_diameter"
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        A look table is provided in textbooks with ratio of
        outerannulus_diameter/innerannulus_diameter in one column and
        Nu_at_innerannulus_diameter and Nu_at_outerannulus_diameter in other
        two columns. This table is implemented and values are interpolated
        as required.
            
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_annulus_laminar(outerannulus_diameter=0.1,
                                    innerannulus_diameter=0.025,
                                    findNUat_ID="innerannulus_diameter")
        741.7512113817678

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        di_do_vals = np.array([0, 0.05, 0.1, 0.25, 0.5, 1])
        Nu_i_vals = np.array([np.nan, 17.46, 11.56, 7.37, 5.74, 4.86])
        Nu_o_vals = np.array([3.66, 4.06, 4.11, 4.23, 4.43, 4.86])
        fxn_Nu_i = interpolate.interp1d(di_do_vals, Nu_i_vals)
        fxn_Nu_o = interpolate.interp1d(di_do_vals, Nu_o_vals)
        if findNUat_ID=="innerannulus_diameter":
            Nu = fxn_Nu_i(innerannulus_diameter/outerannulus_diameter) 
        if findNUat_ID=="outerannulus_diameter":
            Nu = fxn_Nu_o(innerannulus_diameter/outerannulus_diameter)   
        if np.isnan(Nu):
            print("For inner to outer annulus diameter ratio of "
                  "< 0.05, data for 'Nu' is not available")
        else:
            return float(Nu) # Nu is an np.array of float.
                             # Convert to Python float type before returning.
    
    def Nu_liquidmetal_isothermal(self, Re=None, Pr_surface=None):
        r"""Nusselt number for liquid metals (isothermal case).
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number.
        Pr_surface : `int or float`
            Prandtl number at surface temperature.
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 4.8 + 0.0156 Re^{0.85} Pr_{surface}^{0.93}
        
        *where:*
            Re = Reynolds number
            
            :math:`Pr_{surface}` = Prandtl number at surface temperature
            
            :math:`10^4\eqslantless Re \eqslantless 10^6` (for turbulent flow)
            
            :math:`0.004 \eqslantless Pr \eqslantless 0.01`
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                
                except :math:`\Pr_{surface}` is at surface temperature
            
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube. Nu_liquidmetal_isothermal(Re=13570, Pr_surface=0.0119)
        5.624282620227201

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        Nu = 4.8 + 0.0156*np.power(Re,0.85)*np.power(Pr_surface, 0.93)
        if not (1e4<=Re<=1e6): warnings.warn("Re is not within range.")
        return Nu
    
    def Nu_liquidmetal_constflux(self, Re=None, Pr_surface=None):
        r"""Nusselt number for liquid metals (constant flux case).
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number.
        Pr_surface : `int or float`
            Prandtl number at surface temperature.
        

        Returns
        -------
        Nu : `int or float`
            Average Nusselt number.
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            Nu = 6.3 + 0.0167 Re^{0.85} Pr_{surface}^{0.93}
        
        *where:*
            Re = Reynolds number
            
            :math:`Pr_{surface}` = Prandtl number at surface temperature
            
            :math:`10^4\eqslantless Re \eqslantless 10^6` (for turbulent flow)
            
            :math:`0.004 \eqslantless Pr \eqslantless 0.01`
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`
                
                except :math:`\Pr_{surface}` is at surface temperature
            
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.Nu_liquidmetal_constflux(Re=13570, Pr_surface=0.0119)
        7.182405112679119

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        Nu = 6.3 + 0.0167*np.power(Re,0.85)*np.power(Pr_surface, 0.93)
        if not (1e4<=Re<=1e6): warnings.warn("Re is not within range.")
        return Nu
    
    def pressuredrop(self, length=None, diameter=None, frictionfactor=None,
                     density=None, velocity=None):
        r"""Computes pressure drop for flow in a pipe.
        
        Parameters
        ----------
        length : `int or float`
            Length of pipe.
        diameter : `int or float`
            Diameter of pipe.
        frictionfactor: `int or float`
            Friction factor of flowing fluid in pipe.
        density : `int or float`
            Density of fluid.
        velocity : `int or float`
            Velocity of fluid in pipe.
            
        
        Returns
        -------
        pressure drop : `int or float`
            Pressure drop from flowing fluid in pipe
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            \Delta P = f \frac{L}{D} \frac{\rho V_{avg}^2}{2}
        
        *where:*
            
            :math:`\rho` = density
            
            :math:`V_{avg}` = average velocity
            
            L = pipe length
            
            D = pipe inner diameter
            
            f = friction factor
            
            :math:`\Delta P` = pressure drop in pipe

                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`

            
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.pressuredrop(length=200, diameter=0.3, frictionfactor=0.1006,
                         density=888.1, velocity=2)
        119123.81333333332

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        return frictionfactor *length/diameter/2*density*velocity**2
    
    def frictionfactor_laminar(self, Re=None):
        r"""Computes friction factor for laminar flow in a pipe.
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number
        
            
        Returns
        -------
        friction factor : `int or float`
            Friction factor for flowing fluid in pipe
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            f = \frac{64}{Re}
        
        *where:*
            
            Re = Reynolds number
            
            f = friction factor
            
            :math:`Re < 2300` (for laminar flow)
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`

        
        Warnings
        --------
        A friction factor is returned based on the equation
        even if parameters (such as *Re*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.frictionfactor_laminar(636)
        0.10062893081761007

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        if not (Re<2300): warnings.warn("Flow is not laminar."
                                        " Re is not within range.")
        return 64/Re
    
    
    def frictionfactor_turbulent(self, Re=None, diameter=None, 
                                 roughness=None, surfacetype=None):
        r"""Computes friction factor for turbulent flow in a pipe.
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number.
        diameter : `int or float`
            Inner diameter of pipe.
        roughness : `int or float`
            Roughness of pipe.
        surfacetype : `str`
            Indicates if surface is to be considered
            "rough" or "smooth". At least one of the two must be entered.
            
        Returns
        -------
        friction factor : `int or float`
            Friction factor for flowing fluid in pipe
       
        
        Notes
        -----
        The following formula is used:         
        
        for "smooth" surface
            
        .. math::
            
            f=(0.790 \ln Re - 1.64)^{-2} \hspace{5pt}(Petukhov\hspace{5pt}correlation)
            
        for "rough surface"
        
        .. math::
            
            \frac{1}{\sqrt{f}} = ( -2.0 \log_{10} \left( \frac{\epsilon /D}{3.7}
                        + \frac{2.51}{Re \sqrt{f}}\right) 
                                  
                                  \hspace{5pt}(Colebrook\hspace{5pt}correlation)
        
        *where:*
            
            Re = Reynolds number
            
            f = friction factor
            
            :math:`\epsilon` = surface roughness of inside of pipes
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`

        
        Warnings
        --------
        A friction factor is returned based on the equation
        even if parameters (such as *Re*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.frictionfactor_turbulent(Re=126400, diameter=2/12, 
                                     roughness=7e-6, surfacetype="rough")
        0.017397627070796194

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        correctchoice = ["rough", "smooth"]
        if surfacetype not in correctchoice:
            print("You must enter 'surfacetype as 'smooth' OR 'rough'")
            frictionfactor = None
        elif surfacetype == "smooth": # Petukhov equation
            frictionfactor = np.power((0.79*np.log(Re) - 1.64),-2)
        elif surfacetype == "rough":
            def colebrook(f):
                f=f+0.j
                LHS = 1/np.power(f, 0.5)
                RHS = -2*np.log10(roughness/diameter/3.7 + 2.51/Re/np.power(f,0.5))
                return (LHS-RHS).real
            f = fsolve(colebrook, 0.05)
            frictionfactor = f[0]
        if not (Re>2300): warnings.warn("Flow is not turbulent."
                                        " Re is not within range.")
        return frictionfactor
        
    def hydrodynamic_entrylength_laminar(self, Re=None, diameter=None):
        r"""Computes laminar hydrodynamic entry length.
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number.
        diameter : `int or float`
            Inner diameter of pipe.
        
            
        Returns
        -------
        entry length : `int or float`
            Hydrodynamic entry length.
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            L_{h} = 0.05 Re D

        
        *where:*
            
            Re = Reynolds number
            
            D = internal diameter of pipe
            
            :math:`L_h` = hydrodynamic entry length
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`

        
        Warnings
        --------
        A friction factor is returned based on the equation
        even if parameters (such as *Re*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.hydrodynamic_entrylength_laminar(Re=636, diameter=0.3)
        9.54

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        if not (Re<2300): warnings.warn("Flow is not laminar."
                                        " Re is not within range.")
        return 0.05*Re*diameter
    
    def hydrodynamic_entrylength_turbulent(self, diameter=None):
        r"""Computes turbulent hydrodynamic entry length.
        
        Parameters
        ----------
        diameter : `int or float`
            Inner diameter of pipe.
        
            
        Returns
        -------
        entry length : `int or float`
            Hydrodynamic entry length.
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            L_{h} = 10 D

        
        *where:*
            
            D = internal diameter of pipe
            
            :math:`L_h` = hydrodynamic entry length
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.hydrodynamic_entrylength_turbulent(diameter=0.03)
        0.3

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        return 10*diameter
    
    def thermal_entrylength_laminar(self, Re=None, Pr=None, diameter=None):
        r"""Computes laminar thermal entry length.
        
        Parameters
        ----------
        Re : `int or float`
            Reynolds number.
        Pr : `int or float`
            Prandtl number.    
        diameter : `int or float`
            Inner diameter of pipe.
        
            
        Returns
        -------
        entry length : `int or float`
            Thermal entry length.
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            L_{t} = 0.05 Re Pr D

        
        *where:*
            
            Re = Reynolds number
            
            Pr = Prandtl number
            
            D = internal diameter of pipe
            
            :math:`L_t` = thermal entry length
                        
            All fluid properties are at bulk mean temp (:math:`T_{b}`):
            
                :math:`T_{b} = (T_{fluid,in} + T_{fluid,out})/2`

        
        Warnings
        --------
        A friction factor is returned based on the equation
        even if parameters (such as *Re*) do not fall in their
        respective allowable range limits (see above under 'Notes').
        However, if this happens, a warning is issued.
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.thermal_entrylength_laminar(Re=636, Pr=10863, diameter=0.3)
        103633.02

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        if not (Re<2300): warnings.warn("Flow is not laminar."
                                        " Re is not within range.")
        return 0.05*Re*Pr*diameter
    
    def thermal_entrylength_turbulent(self, diameter=None):
        r"""Computes turbulent thermal entry length.
        
        Parameters
        ----------
        diameter : `int or float`
            Inner diameter of pipe.
        
            
        Returns
        -------
        entry length : `int or float`
            Hydrodynamic entry length.
       
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            L_{t} = 10 D

        
        *where:*
            
            D = internal diameter of pipe
            
            :math:`L_t` = thermal entry length
        
        
        Examples
        --------
        First import the module **internalflow**.
           
        >>> from pychemengg.heattransfer import internalflow as intflow 
        >>> tube = intflow.CircularTube()
        >>> tube.thermal_entrylength_turbulent(diameter=0.03)
        0.3

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        """
        return 10*diameter



# -*- coding: utf-8 -*-

# This file is part of pyChemEngg python package.
 
# PyChemEngg: A python-based framework to promote problem solving and critical
# thinking in chemical engineering.

# Copyright (c) 2021 Harvinder Singh Gill <profhsgill@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Module for steady state heat transfer via conduction,
convection, and radiation.

"""
from scipy.optimize import fsolve
from scipy.constants import Stefan_Boltzmann
import math
from matplotlib import pyplot as plt
import numpy as np


__all__ = ["Slab", "Cylinder", "CylinderBase", "Sphere"]

class Slab(object):
    r""" Models a rectangular object


    Parameters
    ----------
    thickness : `int or float`
        Thickness of rectangular slab
    area : `int or float`
        Area of rectangular slab (if area is not known put area = 1.0)
    thermalconductivity : `int or float`
        Thermal conductivity of rectangular slab   
 
    
    Attributes
    ----------
    thickness : `int or float`
        Thickness of rectangular slab
    area : `int or float`
        Area of rectangular slab
    thermalconductivity : `int or float`
        Thermal conductivity of rectangular slab  
    

    Examples
    --------
    First import the module **steadystate**
    
    Units used in this example: SI system
    
    However, any consistent units can be used
    
    >>> from pychemengg.heattransfer import steadystate as ss 
    >>> innerWall = ss.Slab(thickness=0.13, thermalconductivity=0.7, area=1.0)
    # This will create an instance of 'Slab' with a name 'innerwall' 

    """
    def __init__(self, thickness=None, area=None, thermalconductivity=None):
        self.thickness = thickness
        self.area = area
        self.thermalconductivity = thermalconductivity
    
    def heatrateof_cond(self, dT=None):
        r"""Computes heat rate of conduction for a rectangular object
        
        
        Parameters
        ----------
        dT : `int or float`
            Temperature difference between two faces of a slab
                
        
        Returns
        -------
        heatrate : `int or float`
            rate of heat transfer by conduction
        
        
        Notes
        -----
        Heat rate of conduction is calculated using the Fourier's Law
            
        .. math::
            Q (heatrate) = k A \frac{\Delta T}{\Delta x}
        
        *where:*
        
            *k = thermal conductivity*
        
            *A = area of heat transfer*
             
            :math:`\Delta T` *= temperature difference*
             
            :math:`\Delta x` *= slab thickness*
        
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> wall = ss.Slab(thickness=0.13, thermalconductivity=0.7, area=1.0)
        >>> wall.heatrateof_cond(dT=100)
        538.4615384615385
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.       
        """
        return self.thermalconductivity * self.area * dT / self.thickness
    
    def heatrateof_conv(self, heattransfercoefficient=None, dT=None):
        r"""Computes heat rate of convection for a rectangular object

        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the slab surface
        dT : `int or float`
            Temperature difference between slab surface and surrounding fluid
                
        
        Returns
        -------
        heatrate : `int or float`
            Rate of heat transfer by convection
        
        
        Notes
        -----
        Heat rate of convection is calculated using the Newton's Law
            
        .. math::
            Q (heatrate) = h A \Delta T
        
        *where:*
        
            *h = heat transfer coefficient*
        
            *A = area of heat transfer*
             
            :math:`\Delta T` *= temperature difference*
             
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> wall = ss.Slab(thickness=0.13, thermalconductivity=0.7, area=1.0)
        >>> wall.heatrateof_conv(heattransfercoefficient=134.0, dT=60)
        8040.0
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """

        return heattransfercoefficient * self.area * dT
    
    def heatrateof_rad(self, T_infinity=None, T_surface=None, emissivity=None):
        r"""Computes heat rate of radiation for a rectangular object

        
        Parameters
        ----------
        T_infinity : `int or float`
            Temperature of surroundings in **absolute temperature units**
        T_surface : `int or float`
            Temperature of slab surface in **absolute temperature units**
        emissivity : `int or float`
            Emissivity of the slab
        
        
        Returns
        -------
        heatrate : `int or float` (returns a positive value)
            Rate of heat transfer by radiation
        
        
        Notes
        -----
        Heat rate of radiation is calculated using the Stefan-Boltzmann law
            
        .. math::
            Q (heatrate) = \sigma \epsilon A (T_{infinity}^4 - T_{surface}^4)
        
        *where:*
        
            :math:`\sigma` *= Stefan-Boltzmann constant*
            
            :math:`\epsilon` *= emissivity of object*
        
            *A = area of heat transfer*
             
            :math:`T_{infinity}` *= absolute temperature of surroundings*
            
            :math:`T_{surface}` *= absolute temperature of surface*
             
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> wall = ss.Slab(area=1.4)
        >>> wall.heatrateof_rad(T_infinity=10+273, T_surface=30+273, emissivity=0.95)
        151.93639338614008
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """        
        return abs(Stefan_Boltzmann*emissivity*self.area*(T_infinity**4 - T_surface**4))
        
    def resistanceof_cond(self):
        r"""Computes resistance of conduction for a rectangular object

        
        Parameters
        ----------
        `None_required` : 'None'
            Uses attributes defined during instance creation

                
        Returns
        -------
        resistance : `int or float`
            Conduction resistance

               
        Notes
        -----
        The following formula is used:
            
        .. math::
            R_{conduction} = \frac{L}{kA}
        
        *where:*
        
            *L = thickness*
            
            *k = thermal conductivity*
            
            *A = surface area of heat transfer*
            
            :math:`R_{conduction}` *= conduction resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> wall = ss.Slab(thickness=0.3, thermalconductivity=0.9, area=15)
        >>> wall.resistanceof_cond()
        0.022222222222222223
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """        
        return self.thickness / (self.thermalconductivity * self.area)

    def resistanceof_conv(self, heattransfercoefficient=None):
        r"""Computes resistance of convection for a rectangular object

        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the slab surface

        
        Returns
        -------
        resistance : `int or float`
            Convection resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{convection} = \frac{1}{hA}
        
        *where:*
        
            *h = heat transfer coefficient*
                    
            *A = area of heat transfer*
            
            :math:`R_{convection}` *= convection resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> wall = ss.Slab(thickness=0.3, thermalconductivity=0.9, area=0.25*1)
        >>> wall.resistanceof_conv(heattransfercoefficient=10)
        0.4

        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """
        return 1/(heattransfercoefficient * self.area)
    
    def resistanceof_fouling(self, foulingfactor=None):
        r"""Computes resistance of fouling for a rectangular object
        
        
        Parameters
        ----------
        foulingfactor : `int or float`
            Fouling factor :math:`R_f` for the slab surface
            
            typical units are :math:`m^2` K/W        
        
        Returns
        -------
        resistance : `int or float`
            Fouling resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{fouling} = \frac{R_f}{A}
        
        *where:*
        
            :math:`R_f` *= fouling factor*
                    
            *A = fouled area of heat transfer*
            
            :math:`R_{fouling}` *= fouling resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> wall = ss.Slab(thickness=0.3, thermalconductivity=0.9, area=0.25*1)
        >>> wall.resistanceof_fouling(foulingfactor=0.0007)
        0.0028
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """
        return foulingfactor / self.area


class Cylinder(object):
    r""" Models a cylindrical object
    
    
    Parameters
    ----------
    length : `int or float`
        Length of cylindrical object (if length is not known put length = 1.0)
    inner_radius: `int or float`
        Inner radius of cylindrical object
    outer_radius: `int or float`
        Outer radius of cylindrical object    
    thermalconductivity : `int or float`
        Thermal conductivity of cylindrical object   
    
    
    Attributes
    ----------
    length : `int or float`
        Length of cylindrical object (if length is not known put length = 1.0)
    inner_radius: `int or float`
        Inner radius of cylindrical object
    outer_radius: `int or float`
        Outer radius of cylindrical object    
    thermalconductivity : `int or float`
        Thermal conductivity of cylindrical object
    
    
    Examples
    --------
    First import the module **steadystate**
    
    Units used in this example: SI system
    
    However, any consistent units can be used
    
    >>> from pychemengg.heattransfer import steadystate as ss
    >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
    # This will create an instance of 'Cylinder' with a name 'insulation'  
    """
    
    def __init__(self, length=None, inner_radius=None, outer_radius=None, thermalconductivity=None):
        self.length = length
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.thermalconductivity = thermalconductivity
    
    def area(self, radius=None):
        r"""Computes surface area of a cylinder
        
        
        Parameters
        ----------
        radius : `int or float`
            Radius at which surface area is to be computed
                
        
        Returns
        -------
        area : `int or float`
            Surface area of cylinder
        
        
        Notes
        -----
        Surface area of cylinder is computed using:
            
        .. math::
            A = 2\pi rL
                    
        *where:*
        
            *r = radius*
        
            *L = length*
             
            *A = surface area*
        
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss

        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        # This will create an instance of 'Cylinder' with a name 'insulation'
        
        >>> insulation.area(radius=insulation.outer_radius)
        0.10995574287564275
        # This computes surface area @ outer radius
        
        >>> insulation.area(radius=insulation.inner_radius)
        0.047123889803846894
        # This computes surface area @ inner radius
        
        >>> insulation.area(radius=1.5e-3)
        0.047123889803846894
        # This will also compute surface area @ inner radius, which is = 1.5e-3.
        # Here the radius is entered as a number directly, rather than as an attribute.
        """
        return 2 * math.pi * radius * self.length
    
    def volume(self, radius=None):
        r"""Computes volume of a cylinder
        
        
        Parameters
        ----------
        radius : `int or float`
            Radius at which volume is to be computed
                
        
        Returns
        -------
        volume : `int or float`
            Volume of cylinder
        
        
        Notes
        -----
        Volume of cylinder is computed using:
            
        .. math::
            V = \pi r^2 L
                    
        *where:*
        
            *r = radius*
            
            *L = length*
             
            *V = volume*
        
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        This will create an instance of 'Cylinder' with a name 'insulation'
        
        >>> insulation.volume(radius=insulation.outer_radius)
        0.00019242255003237485
        # This computes volume @ outer radius
        
        >>> insulation.volume(radius=insulation.inner_radius)
        3.534291735288517e-05
        # This computes volume @ inner radius
        
        >>>  insulation.volume(radius=1.5e-3)
        3.534291735288517e-05
        # This will also compute volume @ inner radius, which is = 1.5e-3.
        # Here the radius is entered as a number directly, rather than as an attribute.
        """
        return math.pi * radius**2 * self.length
    
    def heatrateof_cond(self, dT=None):
        r"""Computes heat rate of conduction for a cylindrical object
        
        
        Parameters
        ----------
        dT : `int or float`
            Temperature difference between two surfaces of the cylinder
                
        
        Returns
        -------
        heatrate : `int or float`
            Rate of heat transfer by conduction
        
        
        Notes
        -----
        The following formula is used:
            
        .. math::
            Q (heatrate) = \frac{\Delta T}{R_{conduction}}
        
        *where:*
             
            :math:`\Delta T` *= temperature difference*
            
            :math:`R_{conduction}` *= conduction resistance* given by
                         
                :math:`R_{conduction} = \cfrac{ln(r_o/r_i)}{2\pi kL}`
        
                :math:`r_o` *= outer radius of cylinder*
                
                :math:`r_i` *= inner radius of cylinder*
            
                *L = length*
                
                *k = thermal conductivity*
        
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        >>> insulation.heatrateof_cond(dT=75)
        417.12506315941755
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return dT/self.resistanceof_cond()
    
    def heatrateof_conv(self, heattransfercoefficient=None, radius=None, dT=None):
        r"""Computes heat rate of convection for a cylindrical object
        
        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the cylinder surface
        radius : `int or float`
            Radius of cylinder where convective heat transfer rate is to be computed
        dT : `int or float`
            Temperature difference between cylinder surface and surrounding fluid
                
        
        Returns
        -------
        heatrate : `int or float`
            Rate of heat transfer by convection
        
        
        Notes
        -----
        Heat rate of convection is calculated using the Newton's Law
            
        .. math::
            Q (heatrate) = h A \Delta T
        
        *where:*
        
            *h = heat transfer coefficient*
        
            *A = area of heat transfer*
             
            :math:`\Delta T` *= temperature difference*
             
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        >>> insulation.heatrateof_conv(heattransfercoefficient=12, radius=insulation.outer_radius, dT=90.6-30)
        79.9598162191674
        # NOTE: Mathematical expressions (such as: 90.6 - 30) can be used as an argument
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return heattransfercoefficient * self.area(radius=radius) * dT
    
    def heatrateof_rad(self, radius=None, T_infinity=None, T_surface=None, emissivity=None):
        r"""Computes heat rate of radiation for a cylindrical object
        
        
        Parameters
        ----------
        radius : `int or float`
            Radius of cylinder where radiation heat transfer rate is to be computed
        T_infinity : `int or float`
            Temperature of surroundings in **absolute temperature units**
        T_surface : `int or float`
            Temperature of cylinder surface in **absolute temperature units**
        emissivity : `int or float`
            Emissivity of the cylinder
        
        
        Returns
        -------
        heatrate : `int or float` (returns a positive value)
            Rate of heat transfer by radiation
        
        
        Notes
        -----
        Heat rate of radiation is calculated using the Stefan-Boltzmann law
            
        .. math::
            Q (heatrate) = \sigma \epsilon A (T_{infinity}^4 - T_{surface}^4)
        
        *where:*
        
            :math:`\sigma` *= Stefan-Boltzmann constant*
            
            :math:`\epsilon` *= emissivity of object*
        
            *A = area of heat transfer*
             
            :math:`T_{infinity}^4` *= absolute temperature of surroundings*
            
            :math:`T_{surface}^4` *= absolute temperature of surface*
             
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        >>> insulation.heatrateof_rad(radius=insulation.outer_radius, T_infinity=30+273, T_surface=90.6+273, emissivity=0.95)
        53.60018341250309
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return abs(Stefan_Boltzmann * emissivity * self.area(radius=radius) * (T_infinity**4 - T_surface**4))
    
    def resistanceof_cond(self):
        r"""Computes resistance of conduction for a cylindrical object
        
        
        Parameters
        ----------
        `None_required` : 'None'
            Uses attributes defined during instance creation
        
        Returns
        -------
        resistance : `int or float`
            Conduction resistance
                
        Notes
        -----
        The following formula is used:
            
        .. math::
            
            R_{conduction} = \cfrac{ln(r_o/r_i)}{2\pi kL}
        
        *where:*
        
            :math:`r_o` *= outer radius of cylinder*
            
            :math:`r_i` *= inner radius of cylinder*
        
            *L = thickness*
            
            *k = thermal conductivity*
            
            :math:`R_{conduction}` *= conduction resistance*
            
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        >>> insulation.resistanceof_cond()
        0.1798021903357468
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        # Here complex numbers are used.
        # This is because when 'fsolve' is called, say, to find 'thickness' or
        # 'temperature', then this (i.e. resistanceof_cond) method can get called in the process.
        # In that case, 'fsolve', during internal iterations
        # may put 'inner_radius' and 'outer_radius' as negative, 
        # but math.log cannot handle negative numbers. So instead, here np.log is used.
        # But for np.log to handle negative numbers, the argument must be complex, therefore,
        # internal to this method the radii are transfromed to complex numbers. 
        inner_radius = self.inner_radius + 0j
        outer_radius = self.outer_radius + 0j
        value =  (np.log(outer_radius / inner_radius)) / (2 * math.pi * self.thermalconductivity * self.length)
        return value.real # return only real part
    
    def resistanceof_conv(self, heattransfercoefficient=None, radius=None):
        r"""Computes resistance of convection for a cylindrical object
        
        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the cylindrical surface
        radius : `int or float`
            Radius of cylinder where convective heat transfer is to be computed
        
        
        Returns
        -------
        resistance : `int or float`
            Convection resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{convection} = \frac{1}{hA}
        
        *where:*
        
            *h = heat transfer coefficient*
                    
            *A = area of heat transfer*
            
            :math:`R_{convection}` *= convection resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        >>> insulation.resistanceof_conv(heattransfercoefficient=12, radius=insulation.outer_radius)
        0.7578806813899779
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """
        return 1/(heattransfercoefficient * self.area(radius=radius))
    
    def resistanceof_fouling(self, foulingfactor=None, radius=None):
        r"""Computes resistance of fouling for a cylindrical object
        
        
        Parameters
        ----------
        foulingfactor : `int or float`
            Fouling factor :math:`R_f` for the cylindrical surface
            
            typical units are :math:`m^2` K/W        
        
        Returns
        -------
        resistance : `int or float`
            Fouling resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{fouling} = \frac{R_f}{A}
        
        *where:*
        
            :math:`R_f` *= fouling factor*
                    
            *A = fouled area of heat transfer*
            
            :math:`R_{fouling}` *= fouling resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> insulation = Cylinder(length=5, inner_radius=1.5e-3, outer_radius=3.5e-3, thermalconductivity=0.15)
        >>> insulation.resistanceof_fouling(foulingfactor=0.0007, radius=insulation.outer_radius)
        0.006366197723675814
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """        
        return foulingfactor / self.area(radius=radius)
        
    
    
class CylinderBase(object):
    r""" Models the base of a solid circular cylinder
    
    
    Parameters
    ----------
    radius: `int or float`
        Radius of circular base of cylindrical object    
    thermalconductivity : `int or float`
        Thermal conductivity of cylindrical object   
    
    
    Attributes
    ----------
    radius : `int or float`
        Radius of circular base of cylindrical object    
    thermalconductivity : `int or float`
        Thermal conductivity of cylindrical object
    area : `int or float`
        Area of circular base
    
    
    Examples
    --------
    First import the module **steadystate**
    
    Units used in this example: SI system
    
    However, any consistent units can be used
    
    >>> discface = ss.CylinderBase(radius=1.5e-3, thermalconductivity=0.15)
    # This will create an instance of 'CylinderBase' with a name 'discface'  
    """
    
    def __init__(self, radius=None, thermalconductivity=None):
        self.radius = radius
        self.thermalconductivity = thermalconductivity
        self.area =math.pi * self.radius**2
    
    def heatrateof_conv(self, heattransfercoefficient=None, dT=None):
        r"""Computes heat rate of convection for the base of a cylindrical object
        
        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the cylinder base
        radius : `int or float`
            Radius of cylinder base where convective heat transfer rate is to be computed
        dT : `int or float`
            Temperature difference between surface of cylinder base and surrounding fluid
                
        
        Returns
        -------
        heatrate : `int or float`
            Rate of heat transfer by convection
        
        
        Notes
        -----
        Heat rate of convection is calculated using the Newton's Law
            
        .. math::
            Q (heatrate) = h A \Delta T
        
        *where:*
        
            *h = heat transfer coefficient*
        
            *A = area of heat transfer (which is the base circular area)*
             
            :math:`\Delta T` *= temperature difference*
             
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> disc = ss.CylinderBase(radius=0.1, thermalconductivity=401)
        >>> disc.heatrateof_conv(heattransfercoefficient=132, dT=34)
        140.99467829310993
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return heattransfercoefficient * self.area * dT
    
    def heatrateof_rad(self, T_infinity=None, T_surface=None, emissivity=None):
        r"""Computes heat rate of radiation for the base of a cylindrical object
        
        
        Parameters
        ----------
        radius : `int or float`
            Radius of cylinder base where radiation heat transfer rate is to be computed
        T_infinity : `int or float`
            Temperature of surroundings in **absolute temperature units**
        T_surface : `int or float`
            Temperature of cylinder base surface in **absolute temperature units**
        emissivity : `int or float`
            Emissivity of the cylinder base
        
        
        Returns
        -------
        heatrate : `int or float` (returns a positive value)
            Rate of heat transfer by radiation
        
        
        Notes
        -----
        Heat rate of radiation is calculated using the Stefan-Boltzmann law
            
        .. math::
            Q (heatrate) = \sigma \epsilon A (T_{infinity}^4 - T_{surface}^4)
        
        *where:*
        
            :math:`\sigma` *= Stefan-Boltzmann constant*
            
            :math:`\epsilon` *= emissivity of object*
        
            *A = area of heat transfer*
             
            :math:`T_{infinity}^4` *= absolute temperature of surroundings*
            
            :math:`T_{surface}^4` *= absolute temperature of surface*
             
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> disc = ss.CylinderBase(radius=0.1, thermalconductivity=401)
        >>> disc.heatrateof_rad(T_infinity=300, T_surface=550, emissivity=0.9)
        133.72195405218372
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """  
        return abs(Stefan_Boltzmann * emissivity * self.area * (T_infinity**4 - T_surface**4))
    
    def resistanceof_conv(self, heattransfercoefficient=None):
        r"""Computes resistance of convection for the base of a cylindrical object
        
        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the base surface of cylindrical object
        
        
        Returns
        -------
        resistance : `int or float`
            Convection resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{convection} = \frac{1}{hA}
        
        *where:*
        
            *h = heat transfer coefficient*
                    
            *A = area of heat transfer*
            
            :math:`R_{convection}` *= convection resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> disc = ss.CylinderBase(radius=0.1, thermalconductivity=401)
        >>> disc.resistanceof_conv(heattransfercoefficient=132)
        0.24114385316953837
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """
        return 1/(heattransfercoefficient * self.area)
    
    def resistanceof_fouling(self, foulingfactor=None):
        r"""Computes resistance of fouling for the base of cylindrical object
        
        
        Parameters
        ----------
        foulingfactor : `int or float`
            Fouling factor :math:`R_f` for the cylindrical base surface
            
            typical units are :math:`m^2` K/W        
        
        Returns
        -------
        resistance : `int or float`
            Fouling resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{fouling} = \frac{R_f}{A}
        
        *where:*
        
            :math:`R_f` *= fouling factor*
                    
            *A = fouled area of heat transfer*
            
            :math:`R_{fouling}` *= fouling resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> disc = ss.CylinderBase(radius=0.1, thermalconductivity=401)
        >>> disc.resistanceof_fouling(foulingfactor=0.0007)
        0.022281692032865345
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """  
        return foulingfactor / self.area
        
    
class Sphere(object):
    r""" Models a spherical object
    
    
    Parameters
    ----------
    inner_radius: `int or float`
        Inner radius of spherical object
    outer_radius: `int or float`
        Outer radius of spherical object    
    thermalconductivity : `int or float`
        Thermal conductivity of spherical object   
    
    
    Attributes
    ----------
    inner_radius: `int or float`
        Inner radius of spherical object
    outer_radius: `int or float`
        Outer radius of spherical object    
    thermalconductivity : `int or float`
        Thermal conductivity of spherical object 
    
    
    Examples
    --------
    First import the module **steadystate**
    
    Units used in this example: SI system
    
    However, any consistent units can be used
    
    >>> from pychemengg.heattransfer import steadystate as ss
    >>> shellLead = ss.Sphere(inner_radius=0.25, outer_radius=0.30, thermalconductivity=35.3)
    # This will create an instance of 'Sphere' with a name 'shellLead'  
    """
    
    def __init__(self, inner_radius=None, outer_radius=None, thermalconductivity=None):
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.thermalconductivity = thermalconductivity
    
    def area(self, radius=None):
        r"""Computes surface area of a sphere
        
        
        Parameters
        ----------
        radius : `int or float`
            Radius at which surface area is to be computed
                
        
        Returns
        -------
        area : `int or float`
            Surface area of sphere
        
        
        Notes
        -----
        Surface area of sphere is computed using:
            
        .. math::
            A = 4 \pi r^2
                    
        *where:*
        
            *r = radius*
             
            *A = surface area*
        
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        
        >>> shellLead = ss.Sphere(inner_radius=0.25, outer_radius=0.30, thermalconductivity=35.3)
        # This will create an instance of 'Sphere' with a name 'shellLead'
        
        >>> shellLead.area(radius=shellLead.outer_radius)
        1.1309733552923256
        # This computes surface area @ outer radius
        
        >>> shellLead.area(radius=shellLead.inner_radius)
        0.7853981633974483
        # This computes surface area @ inner radius
        
        >>> shellLead.area(radius=0.25)
        0.7853981633974483
        # This will also compute surface area @ inner radius, which is = 0.25
        # Here the radius is entered as a number directly, rather than as an attribute.
        """
        return 4 * math.pi * radius * radius   
    
    def volume(self, radius=None):
        r"""Calculates volume of a sphere
        
        Parameters
        ----------
        radius : `int or float`
            Radius at which volume is to be computed
                
        
        Returns
        -------
        volume : `int or float`
            Volume of sphere
        
        
        Notes
        -----
        Volume of sphere is computed using:
            
        .. math::
            
            V = \frac{4}{3} \pi r^3
                    
        *where:*
        
            *r = radius*
             
            *V = volume*
        
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        
        >>> shellLead = ss.Sphere(inner_radius=0.25, outer_radius=0.30, thermalconductivity=35.3)
        # This will create an instance of 'Sphere' with a name 'shellLead'
        
        >>> shellLead.volume(radius=shellLead.outer_radius)
        0.11309733552923253
        # This computes volume @ outer radius
        
        >>> shellLead.volume(radius=shellLead.inner_radius)
        0.06544984694978735
        # This computes volume @ inner radius
        
        >>>  shellLead.volume(radius=0.25)
        0.06544984694978735
        # This will also compute volume @ inner radius, which is = 0.25.
        # Here the radius is entered as a number directly, rather than as an attribute.
        """
        return 4/3 * math.pi * radius**3
    
    def heatrateof_cond(self, dT=None):
        r"""Computes heat rate of conduction for a spherical object
        
        
        Parameters
        ----------
        dT : `int or float`
            Temperature difference between two surfaces of the sphere
                
        
        Returns
        -------
        heatrate : `int or float`
            Rate of heat transfer by conduction
        
        
        Notes
        -----
        The following formula is used:
            
        .. math::
            Q (heatrate) = \frac{\Delta T}{R_{conduction}}
        
        *where:*
             
            :math:`\Delta T` *= temperature difference*
            
            :math:`R_{conduction}` *= conduction resistance* given by
                         
                :math:`R_{conduction} = \cfrac{(r_o - r_i)}{4\pi k r_o r_i}`
        
                :math:`r_o` *= outer radius of sphere*
                
                :math:`r_i` *= inner radius of sphere*
                
                *k = thermal conductivity*
        
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> shellSS = ss.Sphere(inner_radius=.30, outer_radius=.31, thermalconductivity=15.1)
        >>> shellSS.heatrateof_cond(dT=75)
        132352.15690308428
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return dT/self.resistanceof_cond()
    
    def heatrateof_conv(self, heattransfercoefficient=None, radius=None, dT=None):
        r"""Computes heat rate of convection for a spherical object
        
        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the spherical surface
        radius : `int or float`
            Radius of sphere where convective heat transfer rate is to be computed
        dT : `int or float`
            Temperature difference between spherical surface and surrounding fluid
                
        
        Returns
        -------
        heatrate : `int or float`
            Rate of heat transfer by convection
        
        
        Notes
        -----
        Heat rate of convection is calculated using the Newton's Law
            
        .. math::
            Q (heatrate) = h A \Delta T
        
        *where:*
        
            *h = heat transfer coefficient*
        
            *A = area of heat transfer*
             
            :math:`\Delta T` *= temperature difference*
             
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> shellSS = ss.Sphere(inner_radius=.30, outer_radius=.31, thermalconductivity=15.1)
        >>> shellSS.heatrateof_conv(heattransfercoefficient=500, radius=shellSS.outer_radius, dT=25)
        15095.352700498957
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return heattransfercoefficient * self.area(radius=radius) * dT
    
    def heatrateof_rad(self, radius=None, T_infinity=None, T_surface=None, emissivity=None):
        r"""Computes heat rate of radiation for a spherical object
        
        
        Parameters
        ----------
        radius : `int or float`
            Radius of sphere where radiation heat transfer rate is to be computed
        T_infinity : `int or float`
            Temperature of surroundings in **absolute temperature units**
        T_surface : `int or float`
            Temperature of sphere's surface in **absolute temperature units**
        emissivity : `int or float`
            Emissivity of the sphere
        
        
        Returns
        -------
        heatrate : `int or float` (returns a positive value)
            Rate of heat transfer by radiation
        
        
        Notes
        -----
        Heat rate of radiation is calculated using the Stefan-Boltzmann law
            
        .. math::
            Q (heatrate) = \sigma \epsilon A (T_{infinity}^4 - T_{surface}^4)
        
        *where:*
        
            :math:`\sigma` *= Stefan-Boltzmann constant*
            
            :math:`\epsilon` *= emissivity of object*
        
            *A = area of heat transfer*
             
            :math:`T_{infinity}^4` *= absolute temperature of surroundings*
            
            :math:`T_{surface}^4` *= absolute temperature of surface*
             
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> shellSS = ss.Sphere(inner_radius=.30, outer_radius=.31, thermalconductivity=15.1)
        >>> shellSS.heatrateof_rad(radius=shellSS.outer_radius, T_infinity=30+273, T_surface=90.6+273, emissivity=0.95)
        588.6831572504626
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return abs(Stefan_Boltzmann*emissivity*self.area(radius=radius)*(T_infinity**4 - T_surface**4))

    def resistanceof_cond(self):
        r"""Computes resistance of conduction for a spherical object
        
        
        Parameters
        ----------
        `None_required` : 'None'
            Uses attributes defined during instance creation
        
        Returns
        -------
        resistance : `int or float`
            Conduction resistance
                
        Notes
        -----
        The following formula is used:
            
        .. math::
            
            R_{conduction} = \cfrac{(r_o - r_i)}{4\pi k r_o r_i}
        
        *where:*
        
            :math:`r_o` *= outer radius of sphere*
            
            :math:`r_i` *= inner radius of sphere*
            
            *k = thermal conductivity*
            
            :math:`R_{conduction}` *= conduction resistance*
            
        
        Examples
        --------
        
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> shellSS = ss.Sphere(inner_radius=.30, outer_radius=.31, thermalconductivity=15.1)
        >>> shellSS.resistanceof_cond()
        0.0005666700245385441
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020        
        """
        return (self.outer_radius - self.inner_radius) / (4 * math.pi * self.thermalconductivity * self.outer_radius * self.inner_radius)
    
    def resistanceof_conv(self, heattransfercoefficient=None, radius=None):
        r"""Computes resistance of convection for a spherical object
        
        
        Parameters
        ----------
        heattransfercoefficient : `int or float`
            Heat transfer coefficient *`h`* for the spherical surface
        radius : `int or float`
            Radius of sphere where convective heat transfer is to be computed
        
        
        Returns
        -------
        resistance : `int or float`
            Convection resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{convection} = \frac{1}{hA}
        
        *where:*
        
            *h = heat transfer coefficient*
                    
            *A = area of heat transfer*
            
            :math:`R_{convection}` *= convection resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> shellSS = ss.Sphere(inner_radius=.30, outer_radius=.31, thermalconductivity=15.1)
        >>> shellSS.resistanceof_conv(heattransfercoefficient=500, radius=shellSS.outer_radius)
        0.0016561388459094206
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """
        return 1/(heattransfercoefficient * self.area(radius=radius))

    def resistanceof_fouling(self, foulingfactor=None, radius=None):
        r"""Computes resistance of fouling for a spherical object
        
        
        Parameters
        ----------
        foulingfactor : `int or float`
            Fouling factor :math:`R_f` for the spherical surface
            
            typical units are :math:`m^2` K/W        
        
        Returns
        -------
        resistance : `int or float`
            Fouling resistance
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            R_{fouling} = \frac{R_f}{A}
        
        *where:*
        
            :math:`R_f` *= fouling factor*
                    
            *A = fouled area of heat transfer*
            
            :math:`R_{fouling}` *= fouling resistance*
            
        
        Examples
        --------
        First import the module **steadystate**
        
        Units used in this example: SI system
        
        However, any consistent units can be used
        
        >>> from pychemengg.heattransfer import steadystate as ss
        >>> shellSS = ss.Sphere(inner_radius=.30, outer_radius=.31, thermalconductivity=15.1)
        >>> shellSS.resistanceof_fouling(foulingfactor=0.0007, radius=shellSS.outer_radius)
        0.0005796485960682973
        
        
        References
        ----------
        [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020
        
        """
        return foulingfactor / self.area(radius=radius)



# -*- coding: utf-8 -*-

# This file is part of pyChemEngg python package.
 
# PyChemEngg: A python-based framework to promote problem solving and critical
# thinking in chemical engineering.

# Copyright (c) 2021 Harvinder Singh Gill <profhsgill@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module to compute physical properties of water vapor.

"""

import numpy as np
import os

def _loadwater_data():
    __location__ = os.path.realpath(
                        os.path.join(os.getcwd(), os.path.dirname(__file__)))
    water = []
    file_to_open = "data_waterproperties.txt"
    with open(os.path.join(__location__, file_to_open),"r") as water_file:
        for line in water_file:
            water.append([float(eval(x)) if x != "nan" else np.nan for x in line.strip().split()])
    water_file.close()
    water_data = np.array(water)
    return water_data

def density(T=None):
    r""" Provides density of water vapor at a temperature T.
    
    
    Parameters
    ----------
    T : `int or float`
        Temperature in 'Celsius' at which density is required.


    Returns
    -------
    density : `int or float`
       Density (kg/m3) at temperature T .
   
    
    Notes
    -----
    Look up table adapted from ref [1].
    
    Linear interpolation is performed when the temperature lies
    between tabulated entries.

    
    Examples
    --------
    First import the module **watervaporproperties**.
       
    >>> from pychemengg.physicalproperties import watervaporproperties as wvp 
    >>> wvp.density(T=32.5)
    0.03505

    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    
    """
    water_data = _loadwater_data()
    return np.interp(T, water_data[:,0], water_data[:,3])

def viscosity(T=None):
    r""" Provides viscosity of water vapor at a temperature T.
    
    
    Parameters
    ----------
    T : `int or float`
        Temperature in 'Celsius' at which viscosity is required.


    Returns
    -------
    Viscosity: `int or float`
        Dynamic viscosity (kg/ms) at temperature T.
   
    
    Notes
    -----
    Look up table adapted from ref [1].
    
    Linear interpolation is performed when the temperature lies
    between tabulated entries.

    
    Examples
    --------
    First import the module **watervaporproperties**.
       
    >>> from pychemengg.physicalproperties import watervaporproperties as wvp 
    >>> wvp.viscosity(T=50)
    1.0620000000000002e-05

    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    """
    water_data = _loadwater_data()
    return np.interp(T, water_data[:,0], water_data[:,10])

def specificheat(T=None):
    r""" Provides specific heat of water vapor at a temperature T.
    
    
    Parameters
    ----------
    T : `int or float`
        Temperature in 'Celsius' at which specific heat is required.


    Returns
    -------
    specific heat : `int or float`
        Specific heat (J/kg K) at temperature T.
   
    
    Notes
    -----
    Look up table adapted from ref [1].
    
    Linear interpolation is performed when the temperature lies
    between tabulated entries.

    
    Examples
    --------
    First import the module **watervaporproperties**.
       
    >>> from pychemengg.physicalproperties import watervaporproperties as wvp 
    >>> wvp.specificheat(T=32.5)
    1877.5

    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    
    """
    water_data = _loadwater_data()
    return np.interp(T, water_data[:,0], water_data[:,6])

def thermalconductivity(T=None):
    r""" Provides thermal conductivity of water vapor at a temperature T.
    
    
    Parameters
    ----------
    T : `int or float`
        Temperature in 'Celsius' at which thermal conductivity is required.


    Returns
    -------
    Thermal conductivity : `int or float`
        Thermal conductivity (W/mK) at temperature T.
   
    
    Notes
    -----
    Look up table adapted from ref [1].
    
    Linear interpolation is performed when the temperature lies
    between tabulated entries.

    
    Examples
    --------
    First import the module **watervaporproperties**.
       
    >>> from pychemengg.physicalproperties import watervaporproperties as wvp
    >>> wvp.thermalconductivity(T=42.5)
    0.019799999999999998

    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    
    """
    water_data = _loadwater_data()
    return np.interp(T, water_data[:,0], water_data[:,8])

def heatofcondensation(T=None):
    r""" Provides heat of condensation of water vapor at a temperature T.
    
    
    Parameters
    ----------
    T : `int or float`
        Temperature in 'Celsius' at which heat of condensation is required.


    Returns
    -------
    Heat of condensation : `int or float`
        Heat of condensation (J/kg) at temperature T.
   
    
    Notes
    -----
    Look up table adapted from ref [1].
    
    Linear interpolation is performed when the temperature lies
    between tabulated entries.

    
    Examples
    --------
    First import the module **watervaporproperties**.
       
    >>> from pychemengg.physicalproperties import watervaporproperties as wvp 
    >>> wvp.heatofcondensation(T=68)
    2338800.0
    
    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    
    """
    water_data = _loadwater_data()
    return np.interp(T, water_data[:,0], water_data[:,4])*1e3

def prandtlnumber(T=None):
    r""" Provides Prandtl number of water vapor at a temperature T
    
    
    Parameters
    ----------
    T : `int or float`
        Temperature in 'Celsius' at which Prandtl number is required.


    Returns
    -------
    Prandtl number: `int or float`
        Prandtl number at temperature T.
   
    
    Notes
    -----
    Look up table adapted from ref [1].
    
    Linear interpolation is performed when the temperature lies
    between tabulated entries.

    
    Examples
    --------
    First import the module **watervaporproperties**.
       
    >>> from pychemengg.physicalproperties import watervaporproperties as wvp 
    >>> wvp.prandtlnumber(T=42.5)
    1.0

    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    """
    water_data = _loadwater_data()
    return np.interp(T, water_data[:,0], water_data[:,12])


# col_0 = 'Temp (C)'
# col_1 = 'Saturation pressure (kPa)'
# col_2 = 'Density (liquid, kg/m3)'
# col_3 = 'Density (vapor, kg/m3)'
# col_4 = 'Enthaly/Heat of condensation/condensation  kJ/kg'
# col_5 = 'Specific heat (liquid) J/kg K'
# col_6 = 'Specific heat (vapor) J/kg K'
# col_7 = 'Thermal cond (liquid) W/mK'
# col_8 = 'Thermal cond (vapor) W/mK'
# col_9 = 'Dynamic viscosity (liquid) kg/m s'
# col_10 = 'Dynamic viscosity (vapor) kg/m s'
# col_11 = 'Prandtl Number (liquid)'
# col_12 = 'Prandtl Number (vapor)'
# col_13 = 'Volume expansion coefficient (liquid), beta 1/K'


# -*- coding: utf-8 -*-

# This file is part of pyChemEngg python package.
 
# PyChemEngg: A python-based framework to promote problem solving and critical
# thinking in chemical engineering.

# Copyright (c) 2021 Harvinder Singh Gill <profhsgill@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Calculates fin efficiency 
"""

import math
import scipy.special

__all__ = ["Fin"]


class Fin(object):
    r""" To compute fin efficiency of different fin types.

    Parameters
    ----------
    length : `int or float`
        Length of fin.
    width : `int or float`
        Width of fin.
    thickness : `int or float`
        Thickness of fin.
    diameter : `int or float`
        Diameter of fin.
    inner_radius : `int or float`
        Inner radius of annular fin.
    outer_radius : `int or float`
        Outer radius of annular fin.
    heattransfercoefficient : `int or float`
        Heat transfer coefficient between fin and surroundings.
    thermalconductivity : `int or float`
        Thermal conductivity of fin material.   
 
    
    Attributes
    ----------
    See "Parameters". All parameters are attributes.
    Additional attributes are listed below.
    efficiency : `int or float`
        Fin efficiency.
    surfacearea : `int or float`
        Surface area of given fin geometry.
    
        
    Examples
    --------
    First import the module **fins**.
       
    >>> from heattransfer import fins as fins 
    >>> fin1 = fins.Fin(keyword1=..., keyword2=..., ...)
    # This will create an instance of class 'Fin' and assign it to the
    variable 'fin1'.
    # Methods of the class 'Fin' can then be called like so :-
    # fin1.cylindrical()
    # or
    # fin1.rectangular()

    """
    
    def __init__ (self, length=None, width=None, thickness=None,
                  diameter=None, inner_radius=None, outer_radius=None,
                  heattransfercoefficient=None, thermalconductivity=None):
        self.length = length # fin length
        self.width = width # fin width
        self.thickness = thickness # fin thickness
        self.diameter = diameter # fin diameter
        self.inner_radius = inner_radius # for annular fin: fin inner radius
        self.outer_radius = outer_radius # for annular fin: fin outer radius
        self.heattransfercoefficient = heattransfercoefficient # heat transfer coefficient
        self.thermalconductivity = thermalconductivity # thermal conductivity of fin

    def cylindrical(self):
        r""" Computes fin efficiency for a cylindrical fin.

        Parameters
        ----------
        `None_required` : 'None'
            Attributes that are already defined are used in calculation.
                
        
        Returns
        -------
        tuple containing efficiency and surfacearea :
            (efficiency : `int or float`, surfacearea : `int or float`)
        
        
        Notes
        -----
        The following formula is used:         
        
        .. math::
            
            \eta = \frac {\tanh (mL)} {mL} \\[15pt]
            
            \\[15pt]
            
            A_{surface area} = Perimeter \hspace{2pt} L

        *where:*
            
            L = fin length
            
            m = :math:`\sqrt {\frac {h \hspace{2pt} Perimeter} {k \hspace{2pt} A_{cross-section}}}`
            
            h = heat transfer coefficient
            
            Perimeter = fin perimeter
            
            :math:`A_{cross-section}` = cross section area of fin
            
            :math:`A_{surface area}` = surface area of fin
        
        
        Examples
        --------
        First import the module **fins**.
           
        >>> from pychemengg.heattransfer import fins as fins
        >>> fin1 = fins.Fin(length=0.06,
                             diameter=5e-3,
                             heattransfercoefficient=35,
                             thermalconductivity=8.3)
        >>> fin1.cylindrical()
        (0.2864128219415127, 0.000942477796076938)
        

        References
        ----------
        [1] G. F. Nellis and S. A. Klein, "Introduction to Engineering 
        Heat Transfer", 1st Edition. Cambridge University Press, 2021.
        
        [2] Y. A. Cengel and A. J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        [3] T.L. Bergman, A. S. Lavine, F. P. Incropera, D. P. Dewitt, 
        "Fundamentals of Heat and Mass Transfer", 7th Edition, 
        John Wiley, 2011.
       """
        # def cylindrical(self, length=1, diameter=1, heattransfercoefficient=1, thermalconductivity=1):
        try:
            self.perimeter = math.pi*self.diameter
            self.crosssectionarea = math.pi/4*self.diameter**2 # Area of cross section
            self.surfacearea = self.perimeter * self.length # surface area of fin
            self.mL = math.sqrt(self.heattransfercoefficient * self.perimeter / self.thermalconductivity / self.crosssectionarea) * self.length
            self.efficiency = math.tanh(self.mL)/self.mL
            return self.efficiency, self.surfacearea
        except TypeError:
            print("Check your 'Fin()' object. You maynot have defined all required parameters.")
            print("\ncylindrical fin requires: length=someValue, diameter=someValue,, heattransfercoefficient=someValue, thermalconductivity=someValue")

    def rectangular(self):
        r""" Computes fin efficiency for a rectangular fin.
        

        Parameters
        ----------
        `None_required` : 'None'
            Attributes that are already defined are used in calculation.
                
        
        Returns
        -------
        tuple containing efficiency and surfacearea :
            (efficiency : `int or float`, surfacearea : `int or float`)
        
        
        
        Examples
        --------
        First import the module **fins**.
           
        >>> from pychemengg.heattransfer import fins as fins
        >>> fin1 = fins.Fin (length=0.064, width=1, thickness=0.008, heattransfercoefficient=25, thermalconductivity=210)
        >>> fin1.rectangular()
        (0.9609578814207821, 0.129024)
        

        References
        ----------
        [1] G. F. Nellis and S. A. Klein, "Introduction to Engineering 
        Heat Transfer", 1st Edition. Cambridge University Press, 2021.
        
        [2] Y. A. Cengel and A. J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        [3] T.L. Bergman, A. S. Lavine, F. P. Incropera, D. P. Dewitt, 
        "Fundamentals of Heat and Mass Transfer", 7th Edition, 
        John Wiley, 2011.
       """
        # def rectangular(self, length=1, width=1, thickness=1, heattransfercoefficient=1, thermalconductivity=1):
        try:
            self.perimeter = 2 * (self.width + self.thickness); # perimeter
            self.crosssectionarea = self.width * self.thickness; # Area of cross section
            self.surfacearea = self.perimeter * self.length; # surface area of fin
            self.mL = math.sqrt(self.heattransfercoefficient * self.perimeter / self.thermalconductivity / self.crosssectionarea) * self.length;
            self.efficiency = math.tanh(self.mL)/self.mL;      
            return self.efficiency, self.surfacearea
        except TypeError:
            print("Check your 'Fin()' object. You maynot have defined all required parameters.")
            print("\nrectangular fin requires: length=someValue, width=SomeValue, thickness=someValue, heattransfercoefficient=someValue, thermalconductivity=someValue")

    def straighttriangular(self):
        r""" Computes fin efficiency for a straighttriangular fin.
        

        Parameters
        ----------
        `None_required` : 'None'
            Attributes that are already defined are used in calculation.
                
        
        Returns
        -------
        tuple containing efficiency and surfacearea :
            (efficiency : `int or float`, surfacearea : `int or float`)
        
        
        
        Examples
        --------
        First import the module **fins**.
           
        >>> from pychemengg.heattransfer import fins as fins
        >>> fin1 = fins.Fin(length=0.055, width=0.11, thickness=0.004, heattransfercoefficient=25, thermalconductivity=236)
        >>> fin1.straighttriangular()
        (0.9275971984358753, 0.012107997357118972)
        

        References
        ----------
        [1] G. F. Nellis and S. A. Klein, "Introduction to Engineering 
        Heat Transfer", 1st Edition. Cambridge University Press, 2021.
        
        [2] Y. A. Cengel and A. J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        [3] T.L. Bergman, A. S. Lavine, F. P. Incropera, D. P. Dewitt, 
        "Fundamentals of Heat and Mass Transfer", 7th Edition, 
        John Wiley, 2011.
       """
        # def straighttriangular(self, length=1, width=1, thickness=1, heattransfercoefficient=1, thermalconductivity=1):
        try:
            self.surfacearea = 2 * self.width * math.sqrt(self.length**2 + (self.thickness/2)**2) #surface area of fin
            self.mL = math.sqrt(2 * self.heattransfercoefficient / self.thermalconductivity / self.thickness) * self.length
            self.efficiency = scipy.special.iv(1,2*self.mL) / (self.mL*scipy.special.iv(0,2*self.mL))
            return self.efficiency, self.surfacearea
        except TypeError:
            print("Check your 'Fin()' object. You maynot have defined all required parameters.")
            print("\nstraighttriangular fin requires: length=someValue, width=SomeValue, thickness=someValue, heattransfercoefficient=someValue, thermalconductivity=someValue")

    def straightparabolic(self):
        r""" Computes fin efficiency for a straightparabolic fin.
        

        Parameters
        ----------
        `None_required` : 'None'
            Attributes that are already defined are used in calculation.
                
        
        Returns
        -------
        tuple containing efficiency and surfacearea :
            (efficiency : `int or float`, surfacearea : `int or float`)
        
        
        Examples
        --------
        First import the module **fins**.
           
        >>> from pychemengg.heattransfer import fins as fins
        >>> fin1 = fins.Fin(length=0.03, width=0.2, thickness=0.001, heattransfercoefficient=10, thermalconductivity=125)
        >>> fin1.straightparabolic()
        (0.8867652295764488, 0.012002221851998751)
        

        References
        ----------
        [1] G. F. Nellis and S. A. Klein, "Introduction to Engineering 
        Heat Transfer", 1st Edition. Cambridge University Press, 2021.
        
        [2] Y. A. Cengel and A. J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        [3] T.L. Bergman, A. S. Lavine, F. P. Incropera, D. P. Dewitt, 
        "Fundamentals of Heat and Mass Transfer", 7th Edition, 
        John Wiley, 2011.
       """
    # def straightparabolic(self, length=1, width=1, thickness=1, heattransfercoefficient=1, thermalconductivity=1):
        try:
            self.C1 = math.sqrt(1 + (self.thickness/self.length)**2)
            self.surfacearea = self.width * (self.C1 *self.length + 
                                    self.length**2 / self.thickness * math.log(self.thickness/self.length + self.C1)) #surface area of fin
            self.mL = math.sqrt(2 * self.heattransfercoefficient / self.thermalconductivity / self.thickness) * self.length
            self.efficiency = 2 / (math.sqrt(4*self.mL**2 + 1) + 1)
            return self.efficiency, self.surfacearea
        except:
            print("Check your 'Fin()' object. You maynot have defined all required parameters.")
            print("\nstraightparabolic fin requires: length=someValue, width=SomeValue, thickness=someValue, heattransfercoefficient=someValue, thermalconductivity=someValue")
            

    def pintriangular(self):
        r""" Computes fin efficiency for a pintriangular fin.
        

        Parameters
        ----------
        `None_required` : 'None'
            Attributes that are already defined are used in calculation.
                
        
        Returns
        -------
        tuple containing efficiency and surfacearea :
            (efficiency : `int or float`, surfacearea : `int or float`)
        
                
        Examples
        --------
        First import the module **fins**.
           
        >>> from pychemengg.heattransfer import fins as fins
        >>> fin1 = fins.Fin(length=0.015, diameter=0.0015, heattransfercoefficient=50, thermalconductivity=70)
        fin1.pintriangular()
        >>> fin1.pintriangular()
        (0.9354407776522699, 3.538706842238283e-05)
        

        References
        ----------
        [1] G. F. Nellis and S. A. Klein, "Introduction to Engineering 
        Heat Transfer", 1st Edition. Cambridge University Press, 2021.
        
        [2] Y. A. Cengel and A. J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        [3] T.L. Bergman, A. S. Lavine, F. P. Incropera, D. P. Dewitt, 
        "Fundamentals of Heat and Mass Transfer", 7th Edition, 
        John Wiley, 2011.
       """
    # def pintriangular(self, length=1, diameter=1, heattransfercoefficient=1, thermalconductivity=1):
        try:
            besseli = scipy.special.iv # renaming scipy function for clarity 
            self.surfacearea = math.pi * self.diameter / 2 * math.sqrt(self.length**2 + (self.diameter/2)**2) #surface area of fin
            self.mL = math.sqrt(4*self.heattransfercoefficient/self.thermalconductivity/self.diameter) * self.length
            self.efficiency = 2 * besseli(2, 2*self.mL) / (self.mL*besseli(1,2*self.mL));
            return self.efficiency, self.surfacearea    
        except:
            print("Check your 'Fin()' object. You maynot have defined all required parameters.")
            print("\npintriangular fin requires: length=someValue, diameter=SomeValue, heattransfercoefficient=someValue, thermalconductivity=someValue")
 
    
    def rectangularannular(self):
        r""" Computes fin efficiency for a rectangularannular fin.

        Parameters
        ----------
        `None_required` : 'None'
            Attributes that are already defined are used in calculation.
                
        
        Returns
        -------
        tuple containing efficiency and surfacearea :
            (efficiency : `int or float`, surfacearea : `int or float`)
        
             
        Examples
        --------
        First import the module **fins**.
           
        >>> from pychemengg.heattransfer import fins as fins
        >>> fin1 = fins.Fin(inner_radius=0.025, outer_radius=0.045, thickness=0.001, heattransfercoefficient=50, thermalconductivity=210)
        >>> fin1.rectangularannular()
        (0.922097299521443, 0.00879645943005142)
        

        References
        ----------
        [1] G. F. Nellis and S. A. Klein, "Introduction to Engineering 
        Heat Transfer", 1st Edition. Cambridge University Press, 2021.
        
        [2] Y. A. Cengel and A. J. Ghajar, "Heat And Mass Transfer
        Fundamentals and Applications", 6th Edition. New York, McGraw Hill
        Education, 2020.
        
        [3] T.L. Bergman, A. S. Lavine, F. P. Incropera, D. P. Dewitt, 
        "Fundamentals of Heat and Mass Transfer", 7th Edition, 
        John Wiley, 2011.
       """
    # def rectangularannular(self, inner_radius=1, outer_radius=1, thickness=1, heattransfercoefficient=1, thermalconductivity=1):
        try:
            self.surfacearea = 2 * math.pi * (self.outer_radius**2 - self.inner_radius**2) #surface area of fin
            mro = math.sqrt(2 * self.heattransfercoefficient / self.thermalconductivity / self.thickness) * self.outer_radius
            mri = math.sqrt(2 * self.heattransfercoefficient / self.thermalconductivity / self.thickness) * self.inner_radius
            part1 = 2 * mri/(mro**2 - mri**2)
            besselk = scipy.special.kv # renaming scipy function for clarity
            besseli = scipy.special.iv # renaming scipy function for clarity
            part2 = besselk(1,mri)*besseli(1,mro) - besseli(1,mri)*besselk(1,mro)
            part3 = besseli(0,mri)*besselk(1,mro) + besselk(0,mri)*besseli(1,mro)
            self.efficiency = part1 * part2/part3
            return self.efficiency, self.surfacearea
        except:
            print("Check your 'Fin()' object. You maynot have defined all required parameters.")
            print("\nrectangularannular fin requires: inner_radius=someValue, outer_radius=someValue, thickness=someValue, heattransfercoefficient=someValue, thermalconductivity=someValue")


# -*- coding: utf-8 -*-

# This file is part of pyChemEngg python package.
 
# PyChemEngg: A python-based framework to promote problem solving and critical
# thinking in chemical engineering.

# Copyright (c) 2021 Harvinder Singh Gill <profhsgill@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Module for heat exchanger analysis.

"""

import math
from scipy.optimize import fsolve


__all__ = ["FCorrectionFactor", "EffNTU", "calc_overallheattransfercoefficient_fromNTU"]

class FCorrectionFactor:
    r""" Contains functions to compute 'F' correction factor for LMTD
    for heat exchangers.


    Parameters
    ----------
    `None_required` : 'None'
        This class takes no parameters for instance creation.
 
    
    Attributes
    ----------
    `None_required` : 'None'
        This class does not expose any instance attributes.
    

    Examples
    --------
    First import the module **heatexchangers**.
       
    >>> from pychemengg.heattransfer import heatexchangers as hx 
    >>> hx1 = hx.FCorrectionFactor
    # This will assign the class 'FCorrectionFactor' to the
    # variable 'hx1'.
    # Methods of the class 'FCorrectionFactor' can then be called like so :-
    # hx1.methodname(kwarg1=x, ... etc)

    """
    
    def oneshell2ntubepasses(self, T_tubein=None, T_tubeout=None, T_shellin=None, T_shellout=None):
        r""" To find 'F' correction factor for LMTD.
        Use when shell passes = 1 and tube passes = 2, 4, 6, 8, etc

        
        Parameters
        ----------
        T_tubein : `int or float`
            Temperature of tube side (or cold) fluid at inlet.
        T_tubeout : `int or float`
            Temperature of tube side (or cold) fluid at outlet.
        T_shellin : `int or float`
            Temperature of shell side (or hot) fluid at inlet.
        T_shellout : `int or float`
            Temperature of shell side (or hot) fluid at outlet.        
        

        Returns
        -------
        F : `int or float`
            'F' correction factor for LMTD.
       
        
        Raises
        ------
        ValueError with message 'math domain error'
            If negative number is being passed to the 'log' function
        
        
        Notes
        -----
        Equation number (6) of reference [1] is used.

        This is the same equation that is used to generate plots presented
        in most textbooks.        
        
               
        Examples
        --------
        First import the module **heatexchangers**.
           
        >>> from pychemengg.heattransfer import heatexchangers as hx 
        >>> hx1 = hx.FCorrectionFactor()
        >>> hx1.oneshell2ntubepasses(T_shellin=300, T_shellout=200,
                                     T_tubein=100, T_tubeout=200)
        0.8022781617244771

        

        References
        ----------
        [1] R. A. Bowman, A. C. Mueller, and W. M. Nagle, "Mean Temperature 
        Difference in Design", Transactions of the ASME 62, 1940, pp:283-294.        .

        """
        try:
            R, P = FCorrectionFactor.calc_R_P(self, T_tubein=T_tubein,
                                          T_tubeout=T_tubeout,
                                          T_shellin=T_shellin, 
                                          T_shellout=T_shellout)
            ta = math.sqrt(R**2 + 1)
            tb = 2/P - 1 - R
            if R == 1:
                tt = P / (math.log(10)*(1-P))
            elif R != 1:
                arg1 = (1-P) / (1-P*R)
                tt = 1/(R-1) * math.log10(arg1)
            tc = ta * tt
            td = (tb + ta) / (tb - ta)
            f1_2 = tc/math.log10(td)
            return f1_2
        except ValueError as err:
            if str(err) == "math domain error":
                print(err)
                print("Most likely a negative number is being passed to the"
                      " 'log' function.")
                print("A heat exchanger of this type cannot be designed"
                      " for this particular temperature combination")

        
        
    def twoshell4ntubepasses(self, T_tubein=None, T_tubeout=None, T_shellin=None, T_shellout=None):
        r"""To find 'F' correction factor for LMTD.
        Use when shell passes = 2 and tube passes = 4, 8, 12, etc

        
        Parameters
        ----------
        T_tubein : `int or float`
            Temperature of tube side (or cold) fluid at inlet.
        T_tubeout : `int or float`
            Temperature of tube side (or cold) fluid at outlet.
        T_shellin : `int or float`
            Temperature of shell side (or hot) fluid at inlet.
        T_shellout : `int or float`
            Temperature of shell side (or hot) fluid at outlet.        
        

        Returns
        -------
        F : `int or float`
            'F' correction factor for LMTD.
       
        
        Raises
        ------
        ValueError with message 'math domain error'
            If negative number is being passed to the 'log' function
        
        
        Notes
        -----
        Equation number (8) of reference [1] is used.
        
        This is the same equation used to generate plots presented
        in most textbooks.         
        
               
        Examples
        --------
        First import the module **heatexchangers**.
           
        >>> from pychemengg.heattransfer import heatexchangers as hx 
        >>> hx1 = hx.FCorrectionFactor()
        >>> hx1.twoshell4ntubepasses(T_shellin=300, T_shellout=200,
                                     T_tubein=100, T_tubeout=200)
        0.9568453972970873

        

        References
        ----------
        [1] R. A. Bowman, A. C. Mueller, and W. M. Nagle, "Mean Temperature 
        Difference in Design", Transactions of the ASME 62, 1940, pp:283-294.        .

        """
        try:
            R, P = FCorrectionFactor.calc_R_P(self, T_tubein=T_tubein,
                                              T_tubeout=T_tubeout,
                                              T_shellin=T_shellin,
                                              T_shellout=T_shellout)
            ta = math.sqrt(R**2 + 1)
            tb = 2/P - 1 - R
            tc = 2/P * math.sqrt((1-P) * (1-P*R))
            if R == 1:
                tt = P / (math.log(10)*(1-P))
                tt = tt/2
            elif R != 1:
                arg1 = (1-P) / (1-P*R)
                tt = 1/(R-1) * math.log10(arg1)
                tt = tt/2
            tf = ta * tt
            td = (tb + tc + ta) / (tb + tc - ta)
            f2_2 = tf/math.log10(td)
            return f2_2
        except ValueError as err:
            if str(err) == "math domain error":
                print(err)
                print("Most likely a negative number is being passed to the"
                      " 'log' function.")
                print("A heat exchanger of this type cannot be designed"
                      " for this particular temperature combination")

    
    
    def calc_R_P(self, T_tubein=None, T_tubeout=None, T_shellin=None, T_shellout=None):
        r"""To find 'R' and 'P' parameters needed to compute 'F' correction 
        factors. Could be useful to manually look up tables
        of F correction factors in textbooks.

        
        Parameters
        ----------
        T_tubein : `int or float`
            Temperature of tube side (or cold) fluid at inlet.
        T_tubeout : `int or float`
            Temperature of tube side (or cold) fluid at outlet.
        T_shellin : `int or float`
            Temperature of shell side (or hot) fluid at inlet.
        T_shellout : `int or float`
            Temperature of shell side (or hot) fluid at outlet.        
        

        Returns
        -------
        tuple containg R-parameter and P-parameter :
            (R-parameter : `int or float`, P-parameter : `int or float`)
       
        
        Notes
        -----
        The following formulas defined in reference [1] are used:
        
        .. math::
            
            R = \frac {T_{shellin} - T_{shellout}}  {T_{tubeout} - T_{tubein}}
            
            P = \frac {T_{tubeout} - T_{tubein}} {T_{shellin} - T_{tubein}}
        
               
        Examples
        --------
        First import the module **heatexchangers**.
           
        >>> from pychemengg.heattransfer import heatexchangers as hx 
        >>> hx1 = hx.FCorrectionFactor()
        >>> hx1.calc_R_P(T_shellin=300, T_shellout=200,
                                     T_tubein=100, T_tubeout=200)
        (1.0, 0.5)

        

        References
        ----------
        [1] R. A. Bowman, A. C. Mueller, and W. M. Nagle, "Mean Temperature 
        Difference in Design", Transactions of the ASME 62, 1940, pp:283-294.        .

        """
        R = (T_shellin - T_shellout) / (T_tubeout - T_tubein)
        P = (T_tubeout - T_tubein) / (T_shellin - T_tubein)
        return R, P


    def singlepass_crossflow_bothunmixed(self, T_tubein=None, T_tubeout=None,
                                         T_shellin=None, T_shellout=None):
        r"""To find 'F' correction factor for LMTD.
        Use for single pass cross flow with both fluids unmixed.

        
        Parameters
        ----------
        T_tubein : `int or float`
            Temperature of tube side (or cold) fluid at inlet.
        T_tubeout : `int or float`
            Temperature of tube side (or cold) fluid at outlet.
        T_shellin : `int or float`
            Temperature of shell side (or hot) fluid at inlet.
        T_shellout : `int or float`
            Temperature of shell side (or hot) fluid at outlet.        
        

        Returns
        -------
        F : `int or float`
            'F' correction factor for LMTD.
       
        
        Notes
        -----
        Equation number (10) of reference [1] is used.
        
        This is the same equation that is used to generate plots presented
        in most textbooks.         
        
               
        Examples
        --------
        First import the module **heatexchangers**.
           
        >>> from pychemengg.heattransfer import heatexchangers as hx 
        >>> hx1 = hx.FCorrectionFactor()
        >>> hx1.singlepass_crossflow_bothunmixed(T_shellin=300, T_shellout=200,
                                                 T_tubein=100, T_tubeout=200)
        0.8945911509910063

        

        References
        ----------
        [1] R. A. Bowman, A. C. Mueller, and W. M. Nagle, "Mean Temperature 
        Difference in Design", Transactions of the ASME 62, 1940, pp:283-294.        .

        """
        try:
            T1 = T_shellin
            T2 = T_shellout
            t1 = T_tubein
            t2 = T_tubeout
            p = (T1-T2)/(T1-t1)
            q = (t2-t1)/(T1-t1)
            if p!=q:
                r0 = (p-q)/math.log((1-q)/(1-p)) # r0 stand for countercurrent configuration
            if p==q: # i.e. R = 1
                r0 = (T1-t2)/(T1-t1) # r0 stand for countercurrent configuration
                                     # when R=1 , deltaTm = LMTD = T1-t2 
                                     # so this term is in numerator
            def find_r(r): # this is equation 10 from Bowman, 1940
                n = 40
                summation = 0
                fact = math.factorial
                for u in range(n):
                    for v in range(n):
                        term = (-1)**(u+v) * fact(u+v)/fact(u)/fact(v)/fact(u+1)/fact(v+1)
                        term = term * math.pow(p/r, u) * math.pow(q/r, v)
                        summation = summation + term
                return r-summation
            ans = fsolve(find_r, 0.5)
            F = ans[0]/r0
            # print("p= ",p, " q= ", q, " F= ", F)
            return F
        except ValueError as err:
            if str(err) == "math domain error":
                print(err)
                print("Most likely a negative number is being passed to the"
                      " 'log' function.")
                print("A heat exchanger of this type cannot be designed"
                      " for this particular temperature combination")
    
    def singlepass_crossflow_oneunmixed(self, T_tubein=None, T_tubeout=None,
                                        T_shellin=None, T_shellout=None):
        r"""To find 'F' correction factor for LMTD.
        Use for single pass cross flow with one fluid unmixed.

        
        Parameters
        ----------
        T_tubein : `int or float`
            Temperature of tube side (or cold) fluid at inlet.
        T_tubeout : `int or float`
            Temperature of tube side (or cold) fluid at outlet.
        T_shellin : `int or float`
            Temperature of shell side (or hot) fluid at inlet.
        T_shellout : `int or float`
            Temperature of shell side (or hot) fluid at outlet.        
        

        Returns
        -------
        F : `int or float`
            'F' correction factor for LMTD.
       
        
        Notes
        -----
        Equation number (11) of reference [1] is used.
        
        This is the same equation that is used to generate plots presented
        in most textbooks.         
        
               
        Examples
        --------
        First import the module **heatexchangers**.
           
        >>> from pychemengg.heattransfer import heatexchangers as hx 
        >>> hx1 = hx.FCorrectionFactor()
        >>> hx1.singlepass_crossflow_oneunmixed(T_shellin=300, T_shellout=200,
                                                 T_tubein=100, T_tubeout=200)
        0.8464626304853572

        

        References
        ----------
        [1] R. A. Bowman, A. C. Mueller, and W. M. Nagle, "Mean Temperature 
        Difference in Design", Transactions of the ASME 62, 1940, pp:283-294.        .

        """
        # this is equation 11 from Bowman, 1940
        try:
            T1 = T_shellin
            T2 = T_shellout
            t1 = T_tubein
            t2 = T_tubeout
            P = (T1-T2)/(T1-t1)
            Q = (t2-t1)/(T1-t1)
            if P!=Q:
                r0 = (P-Q)/math.log((1-Q)/(1-P))
            if P==Q:
                r0 = (T1-t2)/(T1-t1)
            b = 1-(Q/P*math.log(1/(1-P)))
            R = Q/math.log(1/(b)) 
            F = R/r0
            return F
        except ValueError as err:
            if str(err) == "math domain error":
                print(err)
                print("Most likely a negative number is being passed to the"
                      " 'log' function.")
                print("A heat exchanger of this type cannot be designed"
                      " for this particular temperature combination")
    
    def singlepass_crossflow_bothmixed(self, T_tubein=None, T_tubeout=None,
                                       T_shellin=None, T_shellout=None):
        r"""To find 'F' correction factor for LMTD.
        Use for single pass cross flow with both fluids mixed.

        
        Parameters
        ----------
        T_tubein : `int or float`
            Temperature of tube side (or cold) fluid at inlet.
        T_tubeout : `int or float`
            Temperature of tube side (or cold) fluid at outlet.
        T_shellin : `int or float`
            Temperature of shell side (or hot) fluid at inlet.
        T_shellout : `int or float`
            Temperature of shell side (or hot) fluid at outlet.        
        

        Returns
        -------
        F : `int or float`
            'F' correction factor for LMTD.
       
        
        Notes
        -----
        Equation number (12) of reference [1] is used.
        
        This is the same equation that is used to generate plots presented
        in most textbooks.         
        
               
        Examples
        --------
        First import the module **heatexchangers**.
           
        >>> from pychemengg.heattransfer import heatexchangers as hx 
        >>> hx1 = hx.FCorrectionFactor()
        >>> hx1.singlepass_crossflow_bothmixed(T_shellin=300, T_shellout=200,
                                                 T_tubein=100, T_tubeout=200)
        0.7959050946318332

        

        References
        ----------
        [1] R. A. Bowman, A. C. Mueller, and W. M. Nagle, "Mean Temperature 
        Difference in Design", Transactions of the ASME 62, 1940, pp:283-294.        .

        """
        # this is equation 12 from Bowman, 1940
        try:
            T1 = T_shellin
            T2 = T_shellout
            t1 = T_tubein
            t2 = T_tubeout
            P = (T1-T2)/(T1-t1)
            Q = (t2-t1)/(T1-t1)
            if P!=Q:
                r0 = (P-Q)/math.log((1-Q)/(1-P))
            if P==Q:
                r0 = (T1-t2)/(T1-t1)
            def findR(R): # this is equation 10 from Bowman, 1940
                ta = (P/R)/(1-math.exp(-P/R))
                tb = (Q/R)/(1-math.exp(-Q/R))
                return R*(ta + tb - 1) - 1
            ans = fsolve(findR, 0.5)
            F = ans[0]/r0
            return F
        except ValueError as err:
            if str(err) == "math domain error":
                print(err)
                print("Most likely a negative number is being passed to the"
                      " 'log' function.")
                print("A heat exchanger of this type cannot be designed"
                      " for this particular temperature combination")


############################
# Following functions need more work before adding to API
############################

    def _find_P_equivalent (self, x, R=None, P=None, shellpass_count=None):
        """ 
            VERIFY AGAIN BEFORE ADDING TO PUBLIC API
        """
        n = shellpass_count
        if R != 1:
            term1 = ((1-x*R) / (1-x))**n
            Pnew = (1-term1) / (R-term1)
            return Pnew-P
        if R == 1:
            Pnew = (x * n) / (x*n - x + 1)
            return Pnew-P 

            
    
        
    def _nshell2ntube(self, T_tubein=None, T_tubeout=None, T_shellin=None,
                     T_shellout=None, shellpass_count=None):
        """ Use when shell passes >= 3
            and tube passes = even multiples of shell pass
            example : 3 shell - 6, 12, 18 ... tube passes
                      4 shell - 8, 12, 16 ... tube passes
                      ... and so on
                      
            VERIFY AGAIN BEFORE ADDING TO PUBLIC API
            
        """
        try:
            n = shellpass_count
            R, P = FCorrectionFactor.calc_R_P(self, T_tubein=T_tubein,
                                              T_tubeout=T_tubeout,
                                              T_shellin=T_shellin,
                                              T_shellout=T_shellout)
            # First we convert P for the 'n' shell '2N' tube exchanger
            # into 'P' that is equivalent to '1' shell '2' tube exchanger                    
            psolved = fsolve(lambda x: FCorrectionFactor._find_P_equivalent(x, R=R, P=P, shellpass_count=n), 0.5)
            pequivalent = psolved[0]
            print("equivalent P =", pequivalent)
            
            # Now we take this P-equivalent and use 1 shell - 2 tube formula to find 'F'
            # To do this we find equivalent T_tubein and T_tubeout values that will correspond 
            # to the exchanger with equivalent P
            # For this we use R, pequivalent, T_shellin and T_shellout and find equivalent T_tubein and equivalent T_tubeout
            equivalenttemps = FCorrectionFactor._get_t1_t2_for_given_P_R (T1=T_shellin, T2=T_shellout, P=pequivalent, R=R)
            t1equivalent = equivalenttemps[2]
            t2equivalent = equivalenttemps[3]
            # Now cal 1-2 exachanger with these equivalent temps
            fn_2n = FCorrectionFactor.oneshell2ntubepasses(T_tubein=t1equivalent,
                                                    T_tubeout=t2equivalent,
                                                    T_shellin=T_shellin,
                                                    T_shellout=T_shellout)
            return fn_2n
        except:
            return "Impossible"

    def _oneshell3tube(self, T_tubein=None, T_tubeout=None, T_shellin=None, T_shellout=None):
        """Based on equation by Fischer
        "Mean Temperature Difference Correction in Multipass Exchangers
        Industrial and Engineering Chemistry, 1938
        vol 30, No 4, page 377-383
        
        VERIFY AGAIN BEFORE ADDING TO PUBLIC API
        
        """
        try:
            R, P = FCorrectionFactor.calc_R_P(self, T_tubein=T_tubein,
                                              T_tubeout=T_tubeout,
                                              T_shellin=T_shellin,
                                              T_shellout=T_shellout)
            R1 = 1/R
            T1 = T_shellin
            T2 = T_shellout
            t1 = T_tubein
            t2 = T_tubeout
            
            if R1 == 1:
                LMTD_countercurrent = abs(T1-t2)
            if R1 !=1:
                LMTD_countercurrent = ((T1-t2)-(T2-t1)) / math.log((T1-t2)/(T2-t1))
           
            def solvefor_thetam(thetaM):
                # thetaM is the true mean temperature from which 'F' can be computed
                # F = thetaM / LMTD_countercurrent
                if R1 != 1:
                    lambda_1 = math.sqrt(9 - 4*R1*(1-R1))
                    m1L =1/2 * (3+lambda_1) * (T1-T2) / (3*thetaM)
                    m2L = 1/2 * (3-lambda_1) * (T1-T2) / (3*thetaM)
                    denom_1 = math.exp(m1L) - math.exp(m2L)
                    A = ((T1-t2)- math.exp(m2L)*(T2-t1)) / denom_1/(1-R1)
                    B = (math.exp(m1L)*(T2-t1) - (T1-t2)) / denom_1/(1-R1)
                    LHS = A*m1L + B*m2L
                    RHS_1 = (T1-T2)/thetaM * (T2 - t1 + math.exp((t2-t1)/3/thetaM) * (T1-t2))
                    RHS_2 = math.exp((t2-t1)/3/thetaM) * (A*m1L*math.exp(m1L) + B*m2L*math.exp(m2L))
                    RHS = RHS_1 - RHS_2
                    return LHS-RHS
                if R1 == 1:
                    # This case was derived by Prof. Harvinder Singh Gill
                    # Equation 12 in the paper by Fischer was adapted by putting R1=1
                    # and the entire solution was reworked for this case
                    # Fischer provided data for R1=1 in the form of a Table (Table - I)
                    # However, Fischer did not show the work nor the equation that
                    # he had used to generate Table I for the case R = 1
                    bLa = 1/9 * (t2-t1) * (t2-T1) / thetaM
                    eal = math.exp((T1-T2)/thetaM)
                    B = (T1-T2+bLa) / (eal-1)
                    A = T2-B
                    termA = math.exp((t2-t1)/3/thetaM) * (3*t2 - 3*T1 - (t2-T1)/3 + 3*B*eal)
                    LHS = (T1-T2) * (3*T2 - 3*t1 - termA)
                    RHS = 3*(T1-T2)*B - 1/3*(t2-t1)*(t2-T1)
                    return LHS-RHS
            ans = fsolve(solvefor_thetam, LMTD_countercurrent)
            F = ans[0]/LMTD_countercurrent
            return F
        except:
            return "Impossible"


    def _nshell3ntube(self, T_tubein=None, T_tubeout=None, T_shellin=None,
                     T_shellout=None, shellpass_count=None):
        """ Use when shell passes >= 2
            and tube passes = multiples of '3' of shell pass
            example : 2 shell-6 tube,
                      3 shell-9 tube,
                      4 shell-12 tube passes
                      ... and so on
                      
            VERIFY AGAIN BEFORE ADDING TO PUBLIC API
            
        """       
        try:
            n = shellpass_count
            R, P = FCorrectionFactor.calc_R_P(self, T_tubein=T_tubein, T_tubeout=T_tubeout, T_shellin=T_shellin, T_shellout=T_shellout)
            R1 = 1/R
            T1 = T_shellin
            T2 = T_shellout
            t1 = T_tubein
            t2 = T_tubeout
            # First we convert P for the 'n' shell '3N' tube exchanger
            # into 'P' that is equivalent to '1' shell '3' tube exchanger           
            psolved = fsolve(FCorrectionFactor._find_P_equivalent, 0.5, R, P, n)
            pequivalent = psolved[0]
            print("equivalent P =", pequivalent)
            
            # Now we take this P-equivalent and use 1 shell - 3 tube formula to find 'F'
            # To do this we find equivalent T_tubein and T_tubeout values that will correspond 
            # to the exchanger with equivalent P
            # For this we use R, pequivalent, T_shellin and T_shellout and find equivalent T_tubein and equivalent T_tubeout
            equivalenttemps = FCorrectionFactor._get_t1_t2_for_given_P_R (T1=T_shellin, T2=T_shellout, P=pequivalent, R=R)
            t1equivalent  = equivalenttemps[2]
            t2equivalent  = equivalenttemps[3]
            # Now cal 1-3 exachanger with these equivalent temps
            fn_2n = FCorrectionFactor._oneshell3tube(T_tubein=t1equivalent, T_tubeout=t2equivalent , T_shellin=T_shellin, T_shellout=T_shellout)
            return fn_2n
        except:
            return "Impossible"

    def _get_t1_t2_for_given_P_R (self, T1=1, T2=1, P=1, R=1):
        t1 = T1-(T1-T2)/P/R
        t2 = P*(T1-t1)+t1
        Rcalc = (T1-T2)/(t2-t1)
        Pcalc = (t2-t1)/(T1-t1)
        return T1,T2,t1,t2,Rcalc,Pcalc


    
    
    def _twopass_crossflow_shellmixed_tubeunmixed(self, T_tubein=None, T_tubeout=None,
                                                 T_shellin=None, T_shellout=None):
        # this is equation 13 from Bowman, 1940
        T1 = T_shellin
        T2 = T_shellout
        t1 = T_tubein
        t2 = T_tubeout
        P = (T1-T2)/(T1-t1)
        Q = (t2-t1)/(T1-t1)
        if P!=Q:
            r0 = (P-Q)/math.log((1-Q)/(1-P))
        if P==Q:
            r0 = (T1-t2)/(T1-t1)
        ta = math.sqrt((1-Q)/(1-P))
        qOverP = Q/P
        tb = 1 - qOverP * math.log((ta-qOverP)/(1-qOverP))
        R = Q/(2* math.log((1)/(tb)))
        F = R/r0
        return F


class EffNTU(object):
    r""" Contains functions to compute 'effectiveness'
    and NTU for heat exchangers.


    Parameters
    ----------
    Cmin : `int or float`
        Minimum heat capacity rate of fluids in heat exchanger.
        :math:`C_{min} = \dot{m} c_p` =  smaller of the two heat 
        capacity rates for the two fluids in the heat exchanger.
            
            :math:`\dot{m}` = mass flow rate of the fluid
            
            :math:`c_p` = specific heat of the fluid
    Cmax : `int or float`
        Maximum heat capacity rate of fluids in heat exchanger.
        :math:`C_{max} = \dot{m} c_p` =  Larger of the two heat l
        capacity rates for the two fluids in the heat exchanger
            
            :math:`\dot{m}` = mass flow rate of the fluid
            
            :math:`c_p` = specific heat of the fluid
    Cratio : `int or float`
        Cmin/Cmax
    effectiveness : `int or float or str`
        Efficiency of the heat exchanger.    
        Input 'int or float' value if effectiveness values is known, or 
        input "?" if it has to be computed. Default = "?".  
    NTU : `int or float or str`
        Number of transfer units for the heat exchanger. 
        Input 'int or float' value if NTU values is known, or 
        input "?" if it has to be computed. Default = "?".
    
    
    Notes
    -----
    1. At least one of effectiveness or NTU must be entered 
    as 'int or float' and the other will get computed.
    
    2. Textbooks contain explicit formulas to compute
    i) effectiveness and ii) NTU for different exchanger types.
    
    3. In this 'class' definition, the following exchanger types
    can be computed with respect to 'effectiveness' or 'NTU''
        
    - Double pipe - parallel flow
    
    - Double pipe - counter flow
    
    - Shell and tube
    
        * one shell pass and 2, 4, 6 etc tube passes
        * n shell passes and 2n, 4n, etc tube passes
        
    - Single pass counter flow:
        
        * both fluids unmixed
        * Cmax fluid mixed, and Cmin unmixed
        * Cmin fluid mixed, and Cmax unmixed
        
    - All exhangers when Cratio = Cmin/Cmax = 0
        
    4. Logic:
        
        Explicit formula for computing 'effectiveness' for each of the above
        exchanger types is implemented in a separate method. 
        When that method is called, it internally calls another
        method ("_get_eff_or_ntu"). This method checks which of
        "effectiveness" or "NTU" is missing. If "effectiveness" is
        missing, and "NTU" is provided, the explict equation is used
        to compute effectiveness from "NTU". If "NTU" is missing, the same
        explict equation for effectiveness becomes an implicit equation for "NTU",
        and it is solved using 'fsolve' of scipy to find "NTU". A
        seperate explict equation for calculating "NTU" from "effectiveness"
        is not implemented. The advantage of this approach is that for the
        case of "Single pass counter flow with both fluid unmixed", there
        is no explict relationship for "NTU" reported in textbooks, but
        there is one for computing "effectiveness". So in this manner by using
        'fsolve', the "NTU"" for the case of "Single pass counter flow
        with both fluid unmixed" can also be computed.
    
    Attributes
    ----------
    See "Parameters". All parameters are attributes.
    

    Examples
    --------
    First import the module **heatexchangers**.
       
    >>> from pychemengg.heattransfer import heatexchangers as hx 
    >>> hx1 = hx.EffNTU(Cmin=0.639, Cmax=0.836, NTU=0.854)
    # This will create an instance of EffNTU and assign it to "hx1".
    # Methods of the class 'EffNTU' can then be called.
    # In this particular example, NTU is assigned = 0.854.
    # By default effectiveness = "?".
    # When a heat exchanger type is next called, it will automatically
    # compute the effectiveness for that exchanger type
    # For example
    >>> hx1.shelltube(shellpass_count=1, tubepass_count=8)
    0.4621565341983249
    # This is the effectiveness.

    """

    def __init__(self, Cmin=None, Cmax=None, NTU="?", effectiveness="?"):
        # C: heat capacity rate = specificheat * massrate
        self.Cmin = Cmin
        self.Cmax = Cmax
        self.Cratio = Cmin/Cmax
        self.effectiveness = effectiveness
        self.NTU = NTU

    def _get_eff_or_ntu(self, func):
        r""" Determines which of "effectiveness" or "NTU" is provided and
        computes the unknown quantity.
        
        
        Parameters
        ----------
        func : `function object`
            function of the form: effectiveness = func(NTU)
        
        
        Returns
        -------
        effectiveness or NTU of heat exchanger : both as `int or float`
            
            
        Notes
        -----
        1. Function - 'func' is passed to this method.
        2. A check is made whether "effectiveness" or "NTU" is unknown.
        3. If "effectiveness" is unknown, the function object, "func"
        is used to compute "effectiveness" from known "NTU" because
        func is explicit for "effectiveness" in the form :
            effectiveness = func(NTU).
        4. If "NTU" is unknown, the same function object, "func" is
        used to solve for "NTU" with known effectiveness by setting
        the equation: func(x) - effectiveness = 0.
        """
        if self.effectiveness == "?" and self.NTU != "?":
            # Case: effectiveness is unknown but NTU is known
            self.effectiveness = func(self.NTU) # computes effectiveness from func
            return self.effectiveness
        if self.NTU == "?" and self.effectiveness != "?":
            # Case: NTU is unknown but effectiveness is known
            guess = 1
            NTUValue = fsolve(lambda x: func(x) - self.effectiveness, guess)
            # solve for NTU using fsolve solver of scipy
            self.NTU = NTUValue[0]
            return self.NTU
        

    def doublepipe_parallelflow(self):
        r""" Computes "effectiveness" or "NTU" for double pipe parallel flow
        heat exchanger.
        """
        def effectiveness(NTU):
            num = (1 - math.exp( - NTU*(1 + self.Cratio)))
            denom = 1 + self.Cratio
            effvalue = num/denom
            return effvalue
        return self._get_eff_or_ntu(effectiveness)
    
    
    def doublepipe_counterflow(self):
        r""" Computes "effectiveness" or "NTU" for double pipe counter flow
        heat exchanger.
        """
        def effectiveness(NTU):
            if self.Cratio < 1:
                num = (1 - math.exp( - NTU*(1 - self.Cratio)))
                denom = (1 - self.Cratio*math.exp( - NTU*(1 - self.Cratio)))
                effvalue = num/denom
                return effvalue
            elif self.Cratio ==1:
                effvalue = NTU/(1+self.NTU)
                return effvalue
        return self._get_eff_or_ntu(effectiveness)
    
    
    def shelltube(self, shellpass_count = None, tubepass_count = None):
        r""" Computes "effectiveness" or "NTU" for shell and tube
        heat exchanger.
        """
        def effectiveness(NTU):
            if tubepass_count%2 == 0:
                self.shellpass_count = shellpass_count
                self.tubepass_count = tubepass_count
                term1 = math.sqrt(1 + self.Cratio**2)
                term2 = math.exp(- NTU * term1)
                b = 2 * (1 + self.Cratio + term1 * (1+term2) / (1-term2))**(-1)
                if self.shellpass_count == 1:
                    effvalue = b
                    return effvalue
                if self.shellpass_count > 1:
                    n = self.shellpass_count
                    if (self.tubepass_count%self.shellpass_count)%2 == 0:
                        c = (1 - b*self.Cratio)/(1 - b)
                        d = (c**n - 1)/(c**n - self.Cratio)
                        effvalue = d
                        return effvalue
                    else:
                        print("Tube pass must be a multiple of shell pass")
                        print("Tube passes must be in multiples of '2'")
                        return None
            else:
                print("Tube passes must be in multiples of '2'")
                return None
        return self._get_eff_or_ntu(effectiveness)


    def crossflow_bothfluids_unmixed(self):
        r""" Computes "effectiveness" or "NTU" for cross flow heat exchanger
        with both fluid unmixed.
        """
        def effectiveness(NTU):
            effvalue = 1-math.exp((1/self.Cratio) 
                                    * NTU**0.22
                                    * (math.exp(- self.Cratio * NTU**0.78)-1))
            return effvalue
        return self._get_eff_or_ntu(effectiveness)            


    def crossflow_Cmin_unmixed(self):
        r""" Computes "effectiveness" or "NTU" for cross flow heat exchanger
        with both fluid with Cmin unmixed.
        """
        def effectiveness(NTU):
            effvalue = (1/self.Cratio) * ( 1-math.exp(- self.Cratio*(1-math.exp(-NTU))))
            return effvalue
        return self._get_eff_or_ntu(effectiveness) 
    
    
    def crossflow_Cmax_unmixed(self):
        r""" Computes "effectiveness" or "NTU" for cross flow heat exchanger
        with both fluid with Cmax unmixed.
        """        
        def effectiveness(NTU):
            effvalue = 1-math.exp(- 1/self.Cratio*(1-math.exp(-self.Cratio*NTU)))
            return effvalue
        return self._get_eff_or_ntu(effectiveness)
    
    
    def Cratio_is_zero(self):
        r""" Computes "effectiveness" or "NTU" for any exchanger with
        Cmin/Cmax = 0
        """
        def effectiveness(NTU):
            effvalue = 1-math.exp(-NTU)
            return effvalue
        return self._get_eff_or_ntu(effectiveness)
    
def calc_overallheattransfercoefficient_fromNTU(NTU=None, area=None, Cmin=None):
    r"""To compute overall heat transfer coefficient (U) from NTU
    
    Parameters
    ----------
    NTU : `int or float`
        Number of transfer units of heat exchanger.
    area : `int or float`
        Surface area of heat transfer for exchanger.
    Cmin: `int or float`
        Minimum heat capacity rate.
        

    Returns
    -------
    Overall heat transfer coefficient (U) : `int or float`
        
    
    
    Notes
    -----
    The following formula is used.
    
    .. math::
        U = \frac {NTU \hspace{2pt} C_{min}} {A_{surfacearea}}
        
           
    *where:*
        
        NTU = number of transfer units for the heat exchanger
        
        :math:`C_{min} = \dot{m} c_p` =  smaller of the two
        capacity rates for the two fluids in the heat exchanger
        
        
        :math:`\dot{m}` = mass flow rate of the fluid
        
        :math:`c_p` = specific heat of the fluid
        
        :math:`A_{surfacearea}` : Heat transfer surface area of the
        exchanger
        
        U = overall heat transfer coefficient
    
    
        
    See Also
    --------
    pychemengg.heattransfer.heatcommonmethods.calc_Ucoefficient
    
    
    Examples
    --------
    First import the module **heatexchangers**.
       
    >>> from pychemengg.heattransfer import heatexchangers as hx 
    >>> hx.calc_overallheattransfercoefficient_fromNTU(NTU=0.651, Cmin=5.02e3, area=5.11)
    639.5342465753424

    
    References
    ----------
    [1] Yunus A. Cengel and Afshin J. Ghajar, "Heat And Mass Transfer
    Fundamentals and Applications", 6th Edition. New York, McGraw Hill
    Education, 2020.
    """
    return NTU * Cmin / area



import setuptools
import os

with open("README.rst", "r") as fh:
    long_description = fh.read()


# Read version number from Version.txt
# __location__ = os.path.realpath(
#                     os.path.join(os.getcwd(), os.path.dirname(__file__)))
# path_to_version_file =  os.path.dirname(os.path.dirname(__location__))
# path_to_version_file = os.path.dirname(os.path.dirname(os.getcwd()))
version_file = "pychemengg/Version.txt"
quotes = ('"', "'")
with open(version_file,"r") as file:
    for line in file:
        if line.startswith('__version__'):
            _, _, current_version = line.split()
            for quote in quotes:
                if current_version.startswith(quote):
                    current_version = current_version.replace(quote,'')            
            break



setuptools.setup(
    name="pychemengg", 
    version=current_version,
    author="Prof. Harvinder Singh Gill",
    author_email="profhsgill@gmail.com",
    description="A framework for problem solving and critical thinking in chemical engineering.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/profhsgill",
    keywords=["chemical engineering", "chemical-engineering", 
              "heat transfer","heat-transfer", 
              "material balances","material-balances",
              "chemical reaction engineering",
              "thermodynamics", "reactions",
              "reactors", "reaction-engineering",
              "process control", "fluid flow"],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",],
    install_requires=["numpy", "scipy"],
    python_requires='>=3',
)